/*
 * (c) 2018-2021 Cloudera, Inc. All rights reserved.
 *
 *   This code is provided to you pursuant to your written agreement with Cloudera, which may be the terms of the
 *  Affero General Public License version 3 (AGPLv3), or pursuant to a written agreement with a third party authorized
 *  to distribute this code.  If you do not have a written agreement with Cloudera or with an authorized and
 *  properly licensed third party, you do not have any rights to this code.
 *
 *   If this code is provided to you under the terms of the AGPLv3:
 *   (A) CLOUDERA PROVIDES THIS CODE TO YOU WITHOUT WARRANTIES OF ANY KIND;
 *   (B) CLOUDERA DISCLAIMS ANY AND ALL EXPRESS AND IMPLIED WARRANTIES WITH RESPECT TO THIS CODE, INCLUDING BUT NOT
 *       LIMITED TO IMPLIED WARRANTIES OF TITLE, NON-INFRINGEMENT, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE;
 *   (C) CLOUDERA IS NOT LIABLE TO YOU, AND WILL NOT DEFEND, INDEMNIFY, OR HOLD YOU HARMLESS FOR ANY CLAIMS ARISING
 *       FROM OR RELATED TO THE CODE; AND
 *   (D) WITH RESPECT TO YOUR EXERCISE OF ANY RIGHTS GRANTED TO YOU FOR THE CODE, CLOUDERA IS NOT LIABLE FOR ANY
 *       DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, PUNITIVE OR CONSEQUENTIAL DAMAGES INCLUDING, BUT NOT LIMITED
 *       TO, DAMAGES RELATED TO LOST REVENUE, LOST PROFITS, LOSS OF INCOME, LOSS OF BUSINESS ADVANTAGE OR
 *       UNAVAILABILITY, OR LOSS OR CORRUPTION OF DATA.
 *
 */

package com.cloudera.naaf.azure.functions;

import com.cloudera.naaf.StandardStatelessNiFiFunction;
import com.cloudera.naaf.StatelessFunctionConfiguration;
import com.cloudera.naaf.StatelessNiFiFunction;
import com.cloudera.naaf.metering.CloudPlatform;
import com.microsoft.azure.functions.ExecutionContext;

import java.io.File;
import java.io.IOException;
import java.util.HashMap;
import java.util.Map;
import java.util.Optional;

public class StatelessNiFiAzureFunctionHolder {
    private static final String UNVERSIONED = "unversioned";

    private static final String EVENT_CONTEXT_PREFIX = "azure.event.context.";

    private static final String FUNCTION_NAME_ENV_VAR = "FUNCTION_NAME";
    private static final String STORAGE_EXTENSIONS_DIRECTORY_ENV_VAR = "STORAGE_EXTENSIONS_DIRECTORY";
    private static final String STORAGE_RESOURCES_DIRECTORY_ENV_VAR = "STORAGE_RESOURCES_DIRECTORY";
    private static final String STORAGE_ENDPOINT_ENV_VAR = "STORAGE_ENDPOINT";
    private static final String STORAGE_CONTAINER_ENV_VAR = "STORAGE_CONTAINER";

    // Defaults
    private static final String DEFAULT_STORAGE_EXTENSIONS_DIRECTORY = "extensions";
    private static final String DEFAULT_STORAGE_RESOURCES_DIRECTORY = "resources";
    private static final String RESOURCE_DOWNLOADER_TYPE = "com.cloudera.naaf.azure.functions.resources.AzureStorageContainerResourceDownloader";
    private static final String STATE_PROVIDER_TYPE = "com.cloudera.naaf.azure.functions.CosmosDBStatelessNiFiStateProvider";
    private static final String REGION = "REGION";

    private static final File NAR_DIRECTORY = new File("lib/");
    private static final File RESOURCES_DIRECTORY = new File(StatelessNiFiFunction.TEMP_DIR, "resources");
    private static final File DEFAULT_EXTENSION_DIRECTORY = new File(StatelessNiFiFunction.TEMP_DIR, "extensions");

    static final String AZURE_CLOUD_PROVIDER = "AZURE";

    private final StatelessNiFiFunction statelessNiFiFunction;

    private final String functionName;

    public StatelessNiFiAzureFunctionHolder() throws IOException {
        functionName = System.getenv(FUNCTION_NAME_ENV_VAR);
        String region = System.getenv(REGION);

        if (functionName == null) {
            throw new IllegalArgumentException(FUNCTION_NAME_ENV_VAR + " environment variable must be specified");
        }

        final File resourcesDirectory = RESOURCES_DIRECTORY;
        if (!resourcesDirectory.exists() && !resourcesDirectory.mkdirs()) {
            throw new IOException("Resources directory " + resourcesDirectory.getAbsolutePath() + " does not exist and cannot be created");
        }

        String resourcesBucketDirectory = null;
        String extensionsBucketDirectory = null;
        final String storageEndpoint = System.getenv(STORAGE_ENDPOINT_ENV_VAR);
        final String storageContainer = System.getenv(STORAGE_CONTAINER_ENV_VAR);

        if (storageEndpoint != null && storageContainer != null) {
            resourcesBucketDirectory = Optional.ofNullable(System.getenv(STORAGE_RESOURCES_DIRECTORY_ENV_VAR))
                    .orElse(DEFAULT_STORAGE_RESOURCES_DIRECTORY);
            extensionsBucketDirectory = Optional.ofNullable(System.getenv(STORAGE_EXTENSIONS_DIRECTORY_ENV_VAR))
                    .orElse(DEFAULT_STORAGE_EXTENSIONS_DIRECTORY);
        }

        final StatelessFunctionConfiguration configurationActions = new StatelessFunctionConfiguration.Builder()
                .cloudPlatform(CloudPlatform.AZURE_FUNCTIONS)
                .resourcesDirectory(resourcesDirectory)
                .resourcesSourceDirectory(resourcesBucketDirectory)
                .extensionsSourceDirectory(extensionsBucketDirectory)
                .defaultExtensionDirectory(DEFAULT_EXTENSION_DIRECTORY)
                .resourceDownloaderType(RESOURCE_DOWNLOADER_TYPE)
                .stateProviderType(STATE_PROVIDER_TYPE)
                .systemClassLoader(this.getClass().getClassLoader())
                .narDirectory(NAR_DIRECTORY)
                .region(region)
                .build();
        statelessNiFiFunction = new StandardStatelessNiFiFunction(functionName, UNVERSIONED, configurationActions);
    }

    public StatelessNiFiFunction getStatelessNiFiFunction() {
        return statelessNiFiFunction;
    }

    static Map<String, String> getAttributes(final ExecutionContext context) {
        final Map<String, String> attributes = new HashMap<>();
        attributes.put(StatelessNiFiFunction.CLOUD_PROVIDER_ATTRIBUTE, StatelessNiFiAzureFunctionHolder.AZURE_CLOUD_PROVIDER);
        attributes.put(EVENT_CONTEXT_PREFIX + "functionName", context.getFunctionName());
        attributes.put(EVENT_CONTEXT_PREFIX + "invocationId", context.getInvocationId());
        if (context.getTraceContext() != null) {
            attributes.put(EVENT_CONTEXT_PREFIX + "trace.parent", context.getTraceContext().getTraceparent());
            attributes.put(EVENT_CONTEXT_PREFIX + "trace.state", context.getTraceContext().getTracestate());
            if (context.getTraceContext().getAttributes() != null) {
                context.getTraceContext().getAttributes().entrySet()
                        .forEach(entry -> attributes.put(EVENT_CONTEXT_PREFIX + ".trace.attributes." + entry.getKey(), entry.getValue()));
            }
        }
        return attributes;
    }

    String getFunctionName() {
        return functionName;
    }
}
