/*
 * (c) 2018-2021 Cloudera, Inc. All rights reserved.
 *
 *   This code is provided to you pursuant to your written agreement with Cloudera, which may be the terms of the
 *  Affero General Public License version 3 (AGPLv3), or pursuant to a written agreement with a third party authorized
 *  to distribute this code.  If you do not have a written agreement with Cloudera or with an authorized and
 *  properly licensed third party, you do not have any rights to this code.
 *
 *   If this code is provided to you under the terms of the AGPLv3:
 *   (A) CLOUDERA PROVIDES THIS CODE TO YOU WITHOUT WARRANTIES OF ANY KIND;
 *   (B) CLOUDERA DISCLAIMS ANY AND ALL EXPRESS AND IMPLIED WARRANTIES WITH RESPECT TO THIS CODE, INCLUDING BUT NOT
 *       LIMITED TO IMPLIED WARRANTIES OF TITLE, NON-INFRINGEMENT, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE;
 *   (C) CLOUDERA IS NOT LIABLE TO YOU, AND WILL NOT DEFEND, INDEMNIFY, OR HOLD YOU HARMLESS FOR ANY CLAIMS ARISING
 *       FROM OR RELATED TO THE CODE; AND
 *   (D) WITH RESPECT TO YOUR EXERCISE OF ANY RIGHTS GRANTED TO YOU FOR THE CODE, CLOUDERA IS NOT LIABLE FOR ANY
 *       DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, PUNITIVE OR CONSEQUENTIAL DAMAGES INCLUDING, BUT NOT LIMITED
 *       TO, DAMAGES RELATED TO LOST REVENUE, LOST PROFITS, LOSS OF INCOME, LOSS OF BUSINESS ADVANTAGE OR
 *       UNAVAILABILITY, OR LOSS OR CORRUPTION OF DATA.
 *
 */

package com.cloudera.naaf.azure.functions.resources;

import com.azure.identity.DefaultAzureCredentialBuilder;
import com.azure.storage.blob.BlobClient;
import com.azure.storage.blob.BlobContainerClient;
import com.azure.storage.blob.BlobServiceClient;
import com.azure.storage.blob.BlobServiceClientBuilder;
import com.azure.storage.blob.models.BlobItem;
import com.azure.storage.blob.models.ListBlobsOptions;
import com.cloudera.naaf.resources.AbstractResourceDownloader;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.OutputStream;
import java.nio.file.Path;
import java.time.Duration;
import java.util.List;
import java.util.stream.Collectors;

public class AzureStorageContainerResourceDownloader extends AbstractResourceDownloader {
    private static final Logger logger = LoggerFactory.getLogger(AzureStorageContainerResourceDownloader.class);

    private static final String STORAGE_ENDPOINT_ENV_VAR = "STORAGE_ENDPOINT";
    private static final String STORAGE_CONTAINER_ENV_VAR = "STORAGE_CONTAINER";

    @Override
    public void download(final String sourceDirectory, final File downloadDirectory) {
        final String storageEndpoint = System.getenv(STORAGE_ENDPOINT_ENV_VAR);
        final String storageContainer = System.getenv(STORAGE_CONTAINER_ENV_VAR);
        logger.info("Downloading {}/{}/{} to {}", storageEndpoint, storageContainer, sourceDirectory, downloadDirectory);
        try {
            final BlobServiceClient blobServiceClient = new BlobServiceClientBuilder()
                    .credential(new DefaultAzureCredentialBuilder().build())
                    .endpoint(storageEndpoint)
                    .buildClient();
            final BlobContainerClient blobContainerClient = blobServiceClient.getBlobContainerClient(storageContainer);
            final ListBlobsOptions options = new ListBlobsOptions();
            final String sourcePrefix = sourceDirectory.endsWith("/") ? sourceDirectory : sourceDirectory + "/";
            options.setPrefix(sourcePrefix);
            final List<String> blobNames = blobContainerClient.listBlobs(options, Duration.ofSeconds(60))
                    .stream().map(BlobItem::getName).collect(Collectors.toList());
            logger.debug(String.format("Successfully listed %s blobs", blobNames.size()));
            for (final String blobName : blobNames) {
                final BlobClient blobClient = blobContainerClient.getBlobClient(blobName);
                final String filenameWithinSourceDir = substringAfter(blobName, sourcePrefix);
                final Path filePath = getDownloadPath(filenameWithinSourceDir, downloadDirectory);
                if (filePath == null) {
                    continue;
                }
                final File file = filePath.toFile();

                logger.info(String.format("Downloading %s to %s", blobName, filePath));
                try (final OutputStream out = new BufferedOutputStream(new FileOutputStream(file))) {
                    blobClient.download(out);
                }
                logger.debug("Downloaded {} ({} bytes)", filePath, file.length());
            }
        } catch (final Exception e) {
            throw new RuntimeException(String.format("Failed to download from %s/%s/%s to %s", storageEndpoint, storageContainer, sourceDirectory,
                    downloadDirectory), e);
        }
    }
}
