/*
 * (c) 2018-2021 Cloudera, Inc. All rights reserved.
 *
 *   This code is provided to you pursuant to your written agreement with Cloudera, which may be the terms of the
 *  Affero General Public License version 3 (AGPLv3), or pursuant to a written agreement with a third party authorized
 *  to distribute this code.  If you do not have a written agreement with Cloudera or with an authorized and
 *  properly licensed third party, you do not have any rights to this code.
 *
 *   If this code is provided to you under the terms of the AGPLv3:
 *   (A) CLOUDERA PROVIDES THIS CODE TO YOU WITHOUT WARRANTIES OF ANY KIND;
 *   (B) CLOUDERA DISCLAIMS ANY AND ALL EXPRESS AND IMPLIED WARRANTIES WITH RESPECT TO THIS CODE, INCLUDING BUT NOT
 *       LIMITED TO IMPLIED WARRANTIES OF TITLE, NON-INFRINGEMENT, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE;
 *   (C) CLOUDERA IS NOT LIABLE TO YOU, AND WILL NOT DEFEND, INDEMNIFY, OR HOLD YOU HARMLESS FOR ANY CLAIMS ARISING
 *       FROM OR RELATED TO THE CODE; AND
 *   (D) WITH RESPECT TO YOUR EXERCISE OF ANY RIGHTS GRANTED TO YOU FOR THE CODE, CLOUDERA IS NOT LIABLE FOR ANY
 *       DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, PUNITIVE OR CONSEQUENTIAL DAMAGES INCLUDING, BUT NOT LIMITED
 *       TO, DAMAGES RELATED TO LOST REVENUE, LOST PROFITS, LOSS OF INCOME, LOSS OF BUSINESS ADVANTAGE OR
 *       UNAVAILABILITY, OR LOSS OR CORRUPTION OF DATA.
 *
 */

package com.cloudera.naaf;

import com.cloudera.naaf.StatelessNiFiStateProvider.StateScope;
import com.cloudera.naaf.metering.Invocation;
import com.cloudera.naaf.metering.MeteringPublisher;
import com.cloudera.naaf.metering.MeteringException;
import com.cloudera.naaf.metering.StandardInvocation;
import com.cloudera.naaf.download.DfxFlowDownload;
import com.cloudera.naaf.download.FlowDownload;
import com.cloudera.naaf.resources.ResourceDownloader;
import org.apache.nifi.components.state.Scope;
import org.apache.nifi.flowfile.FlowFile;
import org.apache.nifi.provenance.ProvenanceEventRecord;
import org.apache.nifi.stateless.bootstrap.StatelessBootstrap;
import org.apache.nifi.stateless.config.ExtensionClientDefinition;
import org.apache.nifi.stateless.config.SslContextDefinition;
import org.apache.nifi.stateless.engine.StatelessEngineConfiguration;
import org.apache.nifi.stateless.flow.DataflowDefinition;
import org.apache.nifi.stateless.flow.DataflowTrigger;
import org.apache.nifi.stateless.flow.StatelessDataflow;
import org.apache.nifi.stateless.flow.TriggerResult;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.ByteArrayInputStream;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.URL;
import java.net.URLClassLoader;
import java.nio.charset.StandardCharsets;
import java.time.Instant;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Optional;
import java.util.Set;
import java.util.UUID;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentMap;
import java.util.concurrent.TimeUnit;
import java.util.stream.Collectors;

public class StandardStatelessNiFiFunction implements StatelessNiFiFunction {
    private static final Logger logger = LoggerFactory.getLogger(StandardStatelessNiFiFunction.class);
    private static final String DEFAULT_DFX_URL = "https://api.us-west-1.cdp.cloudera.com";
    private static final String BILLING_PUBLISHER_CLASS = "com.cloudera.naaf.metering.DbusMeteringPublisher";

    private static final String DEFAULT_SENSITIVE_PROPS_VALUE = "nifi-function"; // This value doesn't really matter. Anything encrypted with this will be deleted when the function terminates
    private static final String DEFAULT_NEXUS_BASE_URL = "https://repository.cloudera.com/artifactory/cloudera-repos/";

    private static final String SNAPSHOT_CONTENTS_PROPERTY_NAME = "nifi.stateless.flow.snapshot.contents";
    private static final String FAILURE_PORT_NAMES_PROPERTY_NAME = "nifi.stateless.failure.port.names";

    private static final String STATELESS_FUNCTION_BOOTSTRAP_LIB = "stateless-function-bootstrap-lib";
    private static final File DEFAULT_WORKING_DIRECTORY = new File(StatelessNiFiFunction.TEMP_DIR, "working");
    private static final String HASH_FILENAME = "nar-digest";
    private static final String NAR_UNPACKED_SUFFIX = "nar-unpacked";

    private final MeteringPublisher meteringPublisher;

    private ClassLoader rootClassLoader;
    private ClassLoader functionBootstrapClassLoader;

    // Mapping of Flow CRN to StatelessDataflow
    private static final ConcurrentMap<String, StatelessDataflow> dataflowMap = new ConcurrentHashMap<>();
    private StatelessDataflow dataflow;
    private String outputPortOfInterest;

    private String flowVersionCrn;
    private String functionId;
    private final String functionName;
    private final String functionVersion;
    private final StatelessFunctionConfiguration statelessFunctionConfiguration;
    private StatelessNiFiStateProvider stateProvider = null;

    public StandardStatelessNiFiFunction(final String functionName, final String functionVersion,
                                         final StatelessFunctionConfiguration statelessFunctionConfiguration) throws IOException {

        if (functionName == null || functionVersion == null) {
            throw new IllegalArgumentException("Function name and version must be provided");
        }

        this.functionName = functionName;
        this.functionVersion = functionVersion;
        this.statelessFunctionConfiguration = statelessFunctionConfiguration;
        this.meteringPublisher = this.instantiateWithBoostrapClassloader(BILLING_PUBLISHER_CLASS);
        initialize();
    }

    @Override
    public TriggerResult trigger(final String instanceId, final InputStream inputStream, final Map<String, String> inputAttributes) {
        final long startNanos = System.nanoTime();  // needed for relative duration calculation with greater than millisecond resolution on Java 8
        final Instant startInstant = Instant.now(); // needed for absolute instant on Calendar time line, without nanosecond resolution
        boolean successfulInvocation = false;

        TriggerResult triggerResult = null;
        try {
            logger.info("Triggering Stateless NiFi Function {} version {} for Dataflow {}", functionName, functionVersion, flowVersionCrn);

            if (stateProvider == null && dataflow.isStateful()) {
                logger.warn("The dataflow stores state, but the StateProvider is not enabled");
            }
            setDataflowState(instanceId, Scope.LOCAL);
            setDataflowState(instanceId, Scope.CLUSTER);

            if (dataflow.getInputPortNames().isEmpty()) {
                logger.info("Ignoring input, since this dataflow has no Input Ports");
            } else {
                enqueueFlowFile(inputStream, inputAttributes);
            }

            final DataflowTrigger trigger = dataflow.trigger();
            triggerResult = trigger.getResult();

            if (triggerResult.isSuccessful()) {
                successfulInvocation = true;
            }

            storeDataflowState(instanceId, Scope.LOCAL);
            storeDataflowState(instanceId, Scope.CLUSTER);

            logger.debug("Returning {} trigger result", successfulInvocation ? "successful" : "failure");
            return triggerResult;
        } catch (final InterruptedException e) {
            Thread.currentThread().interrupt();
            logger.error("Interrupted while running dataflow", e);
            throw new RuntimeException("Interrupted while running dataflow", e);
        } finally {
            final Invocation invocation = StandardInvocation.builder()
                    .durationNanos(System.nanoTime() - startNanos)
                    .startTime(startInstant)
                    .functionId(functionId)
                    .functionName(functionName)
                    .region(statelessFunctionConfiguration.getRegion())
                    .cloudPlatform(statelessFunctionConfiguration.getCloudPlatform())
                    .flowCrn(flowVersionCrn)
                    .successful(successfulInvocation)
                    .build();
            try {
                withFunctionBootstrapClassloader(() -> meteringPublisher.publish(invocation));
            } catch (final MeteringException e) {
                logger.error(e.isInvocationSuccess() ? "Dataflow completed successfully, but failed to publish to metering" :
                        "Dataflow execution failed, and failed to publish to metering", e);
                return new DelegatingTriggerResult(triggerResult, e.isInvocationSuccess(), e);
            }
        }
    }

    @Override
    public void trigger(final String instanceId, final InputStream inputStream, final Map<String, String> inputAttributes, final OutputStream outputStream) throws IOException{
        final TriggerResult triggerResult = trigger(instanceId, inputStream, inputAttributes);
        final FlowFile outputFlowFile = getOutputFlowfile(triggerResult);
        StatelessNiFiUtil.copyOutput(outputFlowFile, triggerResult, outputStream);
        StatelessNiFiUtil.handleTriggerResult(triggerResult);
    }

    private void initialize() throws IOException {
        String dfxUrl = System.getenv(DFX_BASE_URL_VAR);
        if (dfxUrl == null) {
            dfxUrl = DEFAULT_DFX_URL;
        }

        flowVersionCrn = getRequiredEnv(FLOW_VERSION_CRN_VAR);
        final String privateKey = getRequiredEnv(PRIVATE_KEY_VAR);
        final String accessKey = getRequiredEnv(ACCESS_KEY_VAR);

        functionId = new StandardFunctionIdProvider()
                .generateFunctionId(flowVersionCrn, functionName, statelessFunctionConfiguration.getCloudPlatform());

        // AWS Lambda note:
        // When Lambda creates this handler, it will allow 10 seconds for all initialization to complete.
        // After 10 seconds, if the handler has not completed its initialization, Lambda will issue a SIGKILL
        // signal and terminate the process. Then, when the first invocation occurs, it will create a new Handler,
        // and the new Handler instance will have as long as it needs to initialize (or until the Lambda's configured
        // timeout occurs). Lambda behaves this way because the first 10-second initialization phase is free, but
        // the second phase is billed.
        // As a result, we can have a situation where we download some NARs and begin unpacking them but don't complete
        // within the allotted 10 second time interval. The process will be killed with the NAR partially unpacked.
        // When the process is restarted, our NarUnpacker will detect that the NAR has already been expanded and will
        // try to use what is already there. This can result in very confusing failures because not all the files in
        // the NAR were unpacked, so our ClassLoader will only have a portion of the files.  Therefore, we check for
        // the nar-digest file to indicate a completed unpacking, otherwise we delete the directory to unpack again.
        checkWorkingDirectoryIntegrity();

        final StatelessDataflow existingDataflow = dataflowMap.get(flowVersionCrn);
        if (existingDataflow == null) {
            logger.info("Creating dataflow for CRN {}", flowVersionCrn);
            final byte[] dataflowContents = fetchDataflow(dfxUrl, flowVersionCrn, privateKey, accessKey);
            final String dataflowJson = new String(dataflowContents, StandardCharsets.UTF_8);

            logger.debug("Downloaded the following definition for Flow CRN {}:\n{}", flowVersionCrn,dataflowJson);

            final StatelessDataflow createdDataflow = createDataflow(dataflowJson);
            createdDataflow.initialize();
            dataflowMap.put(flowVersionCrn, createdDataflow);
            dataflow = createdDataflow;
        } else {
            logger.info("Dataflow already exists for CRN {}. Will use cached dataflow.", flowVersionCrn);
            dataflow = existingDataflow;
        }

        validatePortNames();

        outputPortOfInterest = getOutputPortOfInterest(dataflow);
        initializeStateProvider();
        withFunctionBootstrapClassloader(() -> meteringPublisher.initialize());
    }

    private void checkWorkingDirectoryIntegrity() {
        final File workingDirectory = getWorkingDirectory();
        purgeIncompleteUnpackedNars(new File(workingDirectory, "nar/extensions"));
        purgeIncompleteUnpackedNars(new File(workingDirectory, "extensions"));
    }

    private void purgeIncompleteUnpackedNars(final File directory) {
        final File[] unpackedDirs = directory.listFiles(file -> file.isDirectory() && file.getName().endsWith(NAR_UNPACKED_SUFFIX));
        if (unpackedDirs == null || unpackedDirs.length == 0) {
            logger.debug("Found no unpacked NARs in {}", directory);
            logger.debug("Directory contains: {}", Arrays.deepToString(directory.listFiles()));
            return;
        }

        for (final File unpackedDir : unpackedDirs) {
            final File narHashFile = new File(unpackedDir, HASH_FILENAME);
            if (!narHashFile.exists()) {
                purgeDirectory(unpackedDir);
            } else {
                logger.debug("Already successfully unpacked {}", unpackedDir);
            }
        }
    }

    private void purgeDirectory(final File directory) {
        if (directory.exists()) {
            deleteRecursively(directory);
            logger.debug("Cleaned up {}", directory);
        }
    }

    private static void deleteRecursively(final File fileOrDirectory) {
        if (fileOrDirectory.isDirectory()) {
            final File[] files = fileOrDirectory.listFiles();
            if (files != null) {
                for (final File file : files) {
                    deleteRecursively(file);
                }
            }
        }

        deleteOrDebug(fileOrDirectory);
    }

    private static void deleteOrDebug(final File file) {
        final boolean deleted = file.delete();
        if (!deleted) {
            logger.debug("Failed to cleanup temporary file {}", file);
        }
    }

    protected String getRequiredEnv(final String environmentVariableName) {
        final String value = System.getenv(environmentVariableName);
        if (value == null) {
            throw new IllegalStateException("Invalid Function Configuration: The " + environmentVariableName + " Environment Variable must be specified");
        }

        return value;
    }

    private void validatePortNames() {
        final String inputPortName = System.getenv(INPUT_PORT_NAME_VAR);
        if (inputPortName != null) {
            final Set<String> inputPorts = dataflow.getInputPortNames();
            if (inputPorts.isEmpty()) {
                throw new IllegalArgumentException(String.format("The %s Environment Variable indicates that the dataflow should enqueue data to the '%s' Input Port, but the configured dataflow " +
                        "has no Input Ports", INPUT_PORT_NAME_VAR));
            }
            if (!inputPorts.contains(inputPortName)) {
                throw new IllegalArgumentException(String.format("The %s Environment Variable indicates that the dataflow should enqueue data to the '%s' Input Port, but the configured dataflow " +
                        "does not have an Input Port with this name. Valid Ports: %s", INPUT_PORT_NAME_VAR, inputPortName, inputPorts));
            }
        } else if (dataflow.getInputPortNames().size() > 1) {
            throw new IllegalArgumentException(String.format("Since the dataflow contains more than one Input Port, the %s Environment Variable is required", INPUT_PORT_NAME_VAR));
        }
    }

    private ClassLoader getRootClassLoader() throws IOException {
        if (rootClassLoader == null) {
            rootClassLoader = createRootClassLoader();
        }
        return rootClassLoader;
    }

    private ClassLoader createRootClassLoader() throws IOException {
        return createURLClassLoader(detectNarDirectory(), StandardStatelessNiFiFunction.class.getClassLoader());
    }

    private ClassLoader createFunctionBootstrapClassLoader(final ClassLoader systemClassLoader) throws IOException {
        return createURLClassLoader(new File(detectNarDirectory(), STATELESS_FUNCTION_BOOTSTRAP_LIB), systemClassLoader);
    }

    private ClassLoader getFunctionBootstrapClassLoader() {
        if (functionBootstrapClassLoader == null) {
            final ClassLoader originalClassLoader = Thread.currentThread().getContextClassLoader();
            final ClassLoader functionSystemClassLoader = statelessFunctionConfiguration.getSystemClassLoader();
            try {
                functionBootstrapClassLoader = functionSystemClassLoader == null ? originalClassLoader :
                        createFunctionBootstrapClassLoader(statelessFunctionConfiguration.getSystemClassLoader());
            } catch (final IOException e) {
                logger.error("Could not create function bootstrap classloader", e);
                throw new RuntimeException(e.getMessage(), e);
            }
        }
        return functionBootstrapClassLoader;
    }

    private ClassLoader createURLClassLoader(final File libraryDirectory, final ClassLoader systemClassLoader) throws IOException {
        final File[] jarFiles = libraryDirectory.listFiles(file -> file.getName().endsWith(".jar"));
        if (jarFiles == null) {
            throw new IOException("Could not access library directory " + libraryDirectory.getAbsolutePath());
        }

        int i=0;
        final URL[] urls = new URL[jarFiles.length];
        for (final File jarFile : jarFiles) {
            final URL url = jarFile.toURI().toURL();
            urls[i++] = url;
        }
        logger.debug("Constructing function bootstrap classloader from URLs: {}", Arrays.asList(urls));
        return new URLClassLoader(urls, systemClassLoader);
    }

    private void initializeStateProvider() {
        final String disableStateProvider = System.getenv(DISABLE_STATE_PROVIDER);
        final boolean disabled = Boolean.TRUE.toString().equalsIgnoreCase(disableStateProvider);
        if (!dataflow.isStateful() || disabled || statelessFunctionConfiguration.getStateProviderType() == null) {
            if (!dataflow.isStateful()) {
                logger.info("Dataflow is not stateful, so the State Provider will not be configured.");
            }
            if (disabled) {
                logger.info("State Provider is disabled.");
            }
            return;
        }

        logger.info("Dataflow is stateful: initializing State Provider");

        this.stateProvider = instantiateWithBoostrapClassloader(statelessFunctionConfiguration.getStateProviderType());
    }

    private void setDataflowState(final String instanceId, final Scope scope) {
        if (stateProvider == null) {
            return;
        }

        if (instanceId == null || scope == null) {
            throw new IllegalArgumentException("Function Instance Identifier and scope are required");
        }
        final StateScope stateScope = StateScope.valueOf(scope.name());
        Map<String, String> state = null;
        try {
            state = getWithFunctionBootstrapClassloader(() -> stateProvider.retrieveState(instanceId, stateScope));
        } catch (final Throwable t) {
            logger.error("Could not retrieve {} state for instance {}", scope, instanceId, t);
        }

        if (state != null) {
            dataflow.setComponentStates(state, scope);
        }
    }

    private void storeDataflowState(final String instanceId, final Scope scope) {
        if (stateProvider == null) {
            return;
        }

        if (instanceId == null || scope == null) {
            throw new IllegalArgumentException("Function Instance Identifier and scope is required");
        }

        final StateScope stateScope = StateScope.valueOf(scope.name());
        final Map<String, String> state = dataflow.getComponentStates(scope);

        withFunctionBootstrapClassloader(() -> stateProvider.storeState(instanceId, stateScope, state));
        try {
            stateProvider.storeState(instanceId, stateScope, state);
        } catch (final Throwable t) {
            logger.error("Could not store {} state for instance {}", scope,  instanceId, t);
        }
    }

    private synchronized void download(final String sourceDirectory, final File downloadDirectory) {
        if (sourceDirectory == null || downloadDirectory == null) {
            return;
        }
        final ResourceDownloader resourceDownloader = instantiateWithBoostrapClassloader(statelessFunctionConfiguration.getResourceDownloaderType());
        try {
            withFunctionBootstrapClassloader(() -> resourceDownloader.download(sourceDirectory, downloadDirectory));
        } catch (final Throwable t) {
            logger.error("Failed to download {}", sourceDirectory, t);
        }
    }

    private StatelessDataflow createDataflow(final String dataflowJson) throws IOException {
        final StatelessEngineConfiguration engineConfiguration = createEngineConfiguration();

        DataflowDefinition dataflowDefinition;
        final StatelessBootstrap bootstrap;
        try {
            final Map<String, String> dataflowDefinitionProperties = new HashMap<>();
            dataflowDefinitionProperties.put(SNAPSHOT_CONTENTS_PROPERTY_NAME, dataflowJson.trim());

            statelessFunctionConfiguration.applyFlowDefinitionConfigAction(dataflowDefinitionProperties);
            if (statelessFunctionConfiguration.getResourcesSourceDirectory() != null) {
                download(statelessFunctionConfiguration.getResourcesSourceDirectory(), statelessFunctionConfiguration.getResourcesDirectory());
            }
            if (statelessFunctionConfiguration.getExtensionsSourceDirectory() != null) {
                download(statelessFunctionConfiguration.getExtensionsSourceDirectory(), engineConfiguration.getExtensionsDirectory());
            }

            final ClassLoader rootClassLoader = getRootClassLoader();
            bootstrap = StatelessBootstrap.bootstrap(engineConfiguration, rootClassLoader);

            dataflowDefinition = bootstrap.parseDataflowDefinition(dataflowDefinitionProperties, Collections.emptyList());

            // Determine the failure ports, if any, and create a new Dataflow Definition using the specified failure ports.
            final Set<String> failurePortNames = getFailurePorts(dataflowDefinition.getOutputPortNames());
            if (!failurePortNames.isEmpty()) {
                final String failurePortString = String.join(",", failurePortNames);
                dataflowDefinitionProperties.put(FAILURE_PORT_NAMES_PROPERTY_NAME, failurePortString);
                dataflowDefinition = bootstrap.parseDataflowDefinition(dataflowDefinitionProperties, Collections.emptyList());
            }

            return bootstrap.createDataflow(dataflowDefinition);
        } catch (final Exception e) {
            throw new RuntimeException("Failed to bootstrap Stateless NiFi Engine", e);
        }
    }

    private String getOutputPortOfInterest(final StatelessDataflow dataflow) {
        final Set<String> outputPortNames = dataflow.getOutputPortNames();
        final Set<String> failurePorts = getFailurePorts(outputPortNames);
        final Set<String> successPorts = new HashSet<>(outputPortNames);
        successPorts.removeAll(failurePorts);

        if (successPorts.isEmpty()) {
            return null;
        }

        if (successPorts.size() > 1) {
            throw new IllegalStateException("Found multiple Output Ports that are not considered failure ports: " + successPorts);
        }

        return successPorts.iterator().next();
    }

    private Set<String> getFailurePorts(final Set<String> outPorts) {
        final String explicitFailurePorts = System.getenv(FAILURE_PORTS_VAR);
        if (explicitFailurePorts != null) {
            return splitCommaSeparated(explicitFailurePorts);
        }

        // If Output Port is explicitly named, all other ports are failure ports.
        final String explicitOutputPort = System.getenv(OUTPUT_PORT_NAME_VAR);
        if (explicitOutputPort != null) {
            final Set<String> portsNotSpecified = new HashSet<>(outPorts);
            portsNotSpecified.remove(explicitOutputPort);
            return portsNotSpecified;
        }

        // Find any ports named 'failure' (ignoring case).
        final Set<String> namedFailure = outPorts.stream()
                .filter(name -> name.trim().equalsIgnoreCase("failure"))
                .collect(Collectors.toSet());

        // Get the set of other ports
        final Set<String> nonFailurePorts = new HashSet<>(outPorts);
        nonFailurePorts.removeAll(namedFailure);

        // If there's 0 or 1 port other than failures, then just use the ports named 'failure' as the failure ports
        if (nonFailurePorts.size() <= 1) {
            return namedFailure;
        }

        throw new IllegalArgumentException("Could not determine which Output Ports are considered successful and which are considered failures. The following ports are ambiguous: " +
                nonFailurePorts + ". Only one Output Port is allowed to be considered successful. The name of the successful Output Port must be set using the Environment Variable " +
                OUTPUT_PORT_NAME_VAR);
    }

    private Set<String> splitCommaSeparated(final String input) {
        final Set<String> splits = new HashSet<>();
        for (final String value : input.split(",")) {
            splits.add(value.trim());
        }

        return splits;
    }

    protected byte[] fetchDataflow(final String dfxBaseUrl, final String flowVersionCrn, final String privateKey, final String accessKey) throws IOException {
        final long start = System.nanoTime();
        try {
            final byte[] flowContents;
            final String flowUrlOverride = System.getenv(FLOW_URL_OVERRIDE);
            if (flowUrlOverride != null && !flowUrlOverride.isEmpty()) {
                flowContents = DfxFlowDownload.fetchDataflowFromUrl(flowUrlOverride);
            } else {
                final FlowDownload download = new DfxFlowDownload(dfxBaseUrl, privateKey, accessKey);
                flowContents = download.downloadFlow(flowVersionCrn);
            }

            final long millis = TimeUnit.NANOSECONDS.toMillis(System.nanoTime() - start);
            logger.info("Successfully downloaded Flow {} in {} millis", flowVersionCrn, millis);

            return flowContents;
        } catch (final Exception e) {
            throw new IOException("Failed to fetch flow from Dataflow Service with provided information", e);
        }
    }

    protected File getWorkingDirectory() {
        final String workingDir = System.getenv(WORKING_DIR);
        final File writableExtensionsDirectory = workingDir == null ? DEFAULT_WORKING_DIRECTORY : new File(workingDir);

        if (!writableExtensionsDirectory.exists() && !writableExtensionsDirectory.mkdirs()) {
            throw new RuntimeException("Extensions directory " + writableExtensionsDirectory.getAbsolutePath() + " does not exist and cannot be created");
        }
        return writableExtensionsDirectory;
    }

    private StatelessEngineConfiguration createEngineConfiguration() throws IOException {
        final File workingDirectory = getWorkingDirectory();
        if (!workingDirectory.exists() && !workingDirectory.mkdirs()) {
            throw new IOException("Working directory " + workingDirectory.getAbsolutePath() + " does not exist and cannot be created");
        }

        final File narDirectory = detectNarDirectory();
        logger.debug("Detected NAR Directory of {}", narDirectory.getAbsolutePath());

        final File contentRepoDir = getContentRepoDirectory();

        final File writableExtensionsDirectory = getWritableExtensionsDirectory();

        final List<File> readOnlyExtensionsDirs = getReadOnlyExtensionDirectories();

        final String krb5EnvVar = System.getenv(KRB5_CONF_FILE);
        final String krb5File = (krb5EnvVar == null) ? DEFAULT_KRB5_CONF_FILE : krb5EnvVar;

        final String processorStartTimeout = System.getenv(PROCESSOR_START_TIMEOUT);
        final String componentEnableTimeout = System.getenv(COMPONENT_ENABLE_TIMEOUT);

        final StatelessEngineConfiguration engineConfiguration = new StatelessEngineConfiguration() {
            @Override
            public File getWorkingDirectory() {
                return StandardStatelessNiFiFunction.this.getWorkingDirectory();
            }

            @Override
            public File getNarDirectory() {
                return narDirectory;
            }

            @Override
            public File getExtensionsDirectory() {
                return writableExtensionsDirectory;
            }

            @Override
            public Collection<File> getReadOnlyExtensionsDirectories() {
                return readOnlyExtensionsDirs;
            }

            @Override
            public File getKrb5File() {
                return new File(krb5File);
            }

            @Override
            public SslContextDefinition getSslContext() {
                return null;
            }

            @Override
            public String getSensitivePropsKey() {
                final String envVar = System.getenv(SENSITIVE_PROPS_VAR);
                return envVar == null ? DEFAULT_SENSITIVE_PROPS_VALUE : envVar;
            }

            @Override
            public List<ExtensionClientDefinition> getExtensionClients() {
                final List<ExtensionClientDefinition> extensionClientDefinitions = new ArrayList<>();

                final String nexusEnvVar = System.getenv(NEXUS_BASE_URL_VAR);
                final String nexusBaseUrl = nexusEnvVar == null ? DEFAULT_NEXUS_BASE_URL : nexusEnvVar;

                final ExtensionClientDefinition definition = new ExtensionClientDefinition();
                definition.setUseSslContext(false);
                definition.setExtensionClientType("nexus");
                definition.setCommsTimeout("30 secs");
                definition.setBaseUrl(nexusBaseUrl);
                extensionClientDefinitions.add(definition);

                return extensionClientDefinitions;
            }

            @Override
            public Optional<File> getContentRepositoryDirectory() {
                return Optional.ofNullable(contentRepoDir);
            }

            @Override
            public String getStatusTaskInterval() {
                return null;
            }

            @Override
            public String getProcessorStartTimeout() {
                if (processorStartTimeout != null) {
                    logger.info("Setting processor start timeout to {}", processorStartTimeout);
                    return processorStartTimeout;
                }
                return StatelessEngineConfiguration.super.getProcessorStartTimeout();
            }

            @Override
            public String getComponentEnableTimeout() {
                if (componentEnableTimeout != null) {
                    logger.info("Setting component enable timeout to {}", componentEnableTimeout);
                    return componentEnableTimeout;
                }
                return StatelessEngineConfiguration.super.getComponentEnableTimeout();
            }
        };

        return engineConfiguration;
    }

    private File getWritableExtensionsDirectory() {
        final String extensionsVar = System.getenv(EXTENSIONS_DOWNLOAD_DIR);
        final File writableExtensionsDirectory = extensionsVar == null ? statelessFunctionConfiguration.getDefaultExtensionDirectory() : new File(extensionsVar);

        if (!writableExtensionsDirectory.exists() && !writableExtensionsDirectory.mkdirs()) {
            throw new RuntimeException("Extensions directory " + writableExtensionsDirectory.getAbsolutePath() + " does not exist and cannot be created");
        }
        return writableExtensionsDirectory;
    }

    private File getContentRepoDirectory() {
        final String contentRepoVar = System.getenv(CONTENT_REPO_LOCATION_VAR);
        if (contentRepoVar == null) {
            return null;
        }

        // Determine the directory name to write content to. This will be made up of the path:
        // {CONTENT_REPO_LOCATION_ENVIRONMENT_VARIABLE}/{function_name}/content_repository/{epoch_timestamp}-{uuid}
        // We use a timestamp in the filename so that it is easy to identify old content that could get left behind if a function is
        // somehow interrupted/terminated ungracefully. We also use a UUID in case there are multiple instances of this function that
        // get run concurrently where the timestamp could conflict.
        final File contentRepoRootDir = new File(contentRepoVar.trim());
        final File functionDir = new File(contentRepoRootDir, functionName);
        final File contentRepositoryDirectory = new File(functionDir, "content_repository");
        final String timestamp = String.valueOf(System.currentTimeMillis());
        final File contentRepoDir = new File(contentRepositoryDirectory, timestamp + "-" + UUID.randomUUID());
        logger.info("Will write FlowFile content to Content Repository directory {}", contentRepoDir);

        return contentRepoDir;
    }

    private List<File> getReadOnlyExtensionDirectories() {
        final List<File> readOnlyExtensionsDirs = new ArrayList<>();
        System.getenv().keySet().stream()
                .filter(key -> key.startsWith(EXTENSIONS_DIR_PREFIX))
                .map(System::getenv)
                .filter(Objects::nonNull)
                .filter(value -> !value.trim().equals(""))
                .map(File::new)
                .forEach(readOnlyExtensionsDirs::add);

        return readOnlyExtensionsDirs;
    }

    private static URLClassLoader getFunctionClassLoader() {
        final ClassLoader classLoader = StandardStatelessNiFiFunction.class.getClassLoader();
        if (!(classLoader instanceof URLClassLoader)) {
            throw new IllegalStateException("Unable to determine the NAR directory");
        }

        return (URLClassLoader) classLoader;
    }

    protected File detectNarDirectory() {
        final File narDirectory = statelessFunctionConfiguration.getNarDirectory();
        if (narDirectory != null && narDirectory.exists()) {
            logger.info("Nar directory is specified as " + narDirectory.getAbsolutePath() + ".  Checking for NARs in it");
            final File[] narFiles = narDirectory.listFiles(f -> f.getName().endsWith(".nar"));
            if (narFiles == null) {
                logger.warn("Found no NARs inside {}.  Attempting further detection of NARs.", narDirectory.getAbsolutePath());
            } else {
                return narDirectory;
            }
        }

        final URLClassLoader urlClassLoader = getFunctionClassLoader();
        final URL[] urls = urlClassLoader.getURLs();
        for (final URL url : urls) {
            final String filename = url.getFile();
            if (filename == null || filename.equals("")) {
                continue;
            }

            final File file = new File(filename);
            if (!file.exists()) {
                logger.warn("Found URL {} in ClassLoader but file {} does not exist", url, file.getAbsolutePath());
                continue;
            }

            if (file.isDirectory()) {
                final File[] narFiles = file.listFiles(f -> f.getName().endsWith(".nar"));
                if (narFiles == null) {
                    logger.warn("Found URL {} in ClassLoader and {} is a directory but cannot access its files", url, file.getAbsolutePath());
                    continue;
                }

                if (narFiles.length > 0) {
                    logger.info("Detected NAR Directory to be {}", file.getAbsolutePath());
                    return file;
                }
            } else if (file.getName().endsWith(".jar")) {
                final File directory = file.getParentFile();
                final File[] narFiles = directory.listFiles(f -> f.getName().endsWith(".nar"));
                if (narFiles == null) {
                    logger.warn("Found URL {} in ClassLoader but cannot access the files in its parent directory {}", url, directory);
                    continue;
                }

                if (narFiles.length > 0) {
                    logger.debug("Detected NAR Directory to be {}", directory.getAbsolutePath());
                    return directory;
                }
            }
        }

        logger.error("ClassLoader that loaded the function did not contain nifi-stateless-bootstrap. URLs that were present: {}", Arrays.asList(urlClassLoader.getURLs()));
        throw new IllegalStateException("Unable to determine the NAR directory");
    }

    @Override
    public void cleanUp() {

    }

    @Override
    public FlowFile getOutputFlowfile(final TriggerResult triggerResult) {
        List<FlowFile> flowFiles;
        if (outputPortOfInterest == null) {
            flowFiles = Collections.emptyList();
        } else {
            flowFiles = triggerResult.getOutputFlowFiles(outputPortOfInterest);
            if (flowFiles == null) {
                flowFiles = Collections.emptyList();
            }
        }

        if (flowFiles.size() > 1) {
            logger.info("Triggered dataflow but it returned multiple FlowFiles. Will not return any results from function: {}", triggerResult.getOutputFlowFiles());
            return null;
        } else if (flowFiles.size() == 1) {
            final FlowFile flowFile = flowFiles.get(0);
            logger.info("Successfully triggered dataflow. Returning {} bytes", flowFile.getSize());
            return flowFile;
        } else {
            logger.info("Successfully triggered dataflow but there were no output FlowFiles. Will not return any result from function.");
            return null;
        }
    }

    private void enqueueFlowFile(final InputStream flowFileContents, final Map<String, String> inputAttributes) {
        final InputStream dataflowInput = flowFileContents == null ? new ByteArrayInputStream(new byte[0]) : flowFileContents;
        final String inputPortName = getInputPortName();
        final Map<String, String> attributes = inputAttributes == null ? Collections.emptyMap() : inputAttributes;
        dataflow.enqueue(dataflowInput, attributes, inputPortName);
        logger.info("Enqueued FlowFile for Input Port {}", inputPortName);
    }

    private String getInputPortName() {
        final String inputPortNameVariable = System.getenv(INPUT_PORT_NAME_VAR);
        if (inputPortNameVariable != null) {
            return inputPortNameVariable;
        }

        final Set<String> inputPortNames = dataflow.getInputPortNames();

        if (inputPortNames.isEmpty()) {
            throw new RuntimeException("At least one Input Port is required in the dataflow");
        } else if (inputPortNames.size() == 1) {
            final String portName = dataflow.getInputPortNames().iterator().next();
            return portName;
        } else {
            throw new RuntimeException("The configured dataflow has " + inputPortNames.size() + " Input Ports, and there is no " + INPUT_PORT_NAME_VAR + " Environment Variable set to indicate " +
                    "which one to use. This function needs to be configured with the '" + INPUT_PORT_NAME_VAR + "' Environment Variable.");
        }
    }

    private <T> T getWithFunctionBootstrapClassloader(final FrameworkAction<T> action) {
        final ClassLoader functionBootstrapClassLoader = getFunctionBootstrapClassLoader();
        final ClassLoader initialClassLoader = Thread.currentThread().getContextClassLoader();
        try {
            Thread.currentThread().setContextClassLoader(functionBootstrapClassLoader);
            return action.execute();
        } finally {
            Thread.currentThread().setContextClassLoader(initialClassLoader);
        }
    }

    private void withFunctionBootstrapClassloader(final Runnable action) {
        getWithFunctionBootstrapClassloader(() -> {
            action.run();
            return null;
        });
    }

    private <T> T instantiateWithBoostrapClassloader(final String className) {
        final ClassLoader initialClassLoader = Thread.currentThread().getContextClassLoader();
        final ClassLoader functionBootstrapClassLoader = getFunctionBootstrapClassLoader();
        try {
            final Class<?> rawClass = Class.forName(className, true, functionBootstrapClassLoader);
            Thread.currentThread().setContextClassLoader(functionBootstrapClassLoader);
            return (T) rawClass.newInstance();
        } catch (final Throwable t) {
            logger.error("Could not instantiate " + className, t);
            return null;
        } finally {
            Thread.currentThread().setContextClassLoader(initialClassLoader);
        }
    }

    private interface FrameworkAction<T> {
        T execute();
    }

    private class DelegatingTriggerResult implements TriggerResult {
        private final TriggerResult delegate;
        private final boolean successful;
        private final Throwable failureCause;

        /**
         * Delegates methods to another trigger result.
         * @param triggerResult An underlying trigger result.  If null, returns the successful and failureCause values provided
         *                      when those methods are called
         * @param successful Provides a successful value if the trigger result is not provided
         * @param failureCauseOverride Provides a failure cause that will override the delegate's failure cause
         */
        DelegatingTriggerResult(final TriggerResult triggerResult, final boolean successful, final Throwable failureCauseOverride) {
            this.delegate = triggerResult;
            this.successful = successful;
            this.failureCause = failureCauseOverride;
        }

        @Override
        public boolean isSuccessful() {
            return delegate == null ? successful : delegate.isSuccessful();
        }

        @Override
        public boolean isCanceled() {
            return delegate == null ? false : delegate.isCanceled();
        }

        @Override
        public Optional<Throwable> getFailureCause() {
            return failureCause != null ? Optional.of(failureCause) : delegate.getFailureCause();
        }

        @Override
        public Map<String, List<FlowFile>> getOutputFlowFiles() {
            return delegate == null ? null : delegate.getOutputFlowFiles();
        }

        @Override
        public List<FlowFile> getOutputFlowFiles(String s) {
            return delegate == null ? null : delegate.getOutputFlowFiles(s);
        }

        @Override
        public InputStream readContent(FlowFile flowFile) throws IOException {
            return delegate == null ? null : delegate.readContent(flowFile);
        }

        @Override
        public byte[] readContentAsByteArray(FlowFile flowFile) throws IOException {
            return delegate == null ? null : delegate.readContentAsByteArray(flowFile);
        }

        @Override
        public void acknowledge() {
            if (delegate != null) {
                delegate.acknowledge();
            }
        }

        @Override
        public void abort(Throwable throwable) {
            if (delegate != null) {
                delegate.abort(throwable);
            }
        }

        @Override
        public List<ProvenanceEventRecord> getProvenanceEvents() throws IOException {
            return delegate == null ? null : delegate.getProvenanceEvents();
        }
    }
}
