/*
 * (c) 2018-2022 Cloudera, Inc. All rights reserved.
 *
 *   This code is provided to you pursuant to your written agreement with Cloudera, which may be the terms of the
 *  Affero General Public License version 3 (AGPLv3), or pursuant to a written agreement with a third party authorized
 *  to distribute this code.  If you do not have a written agreement with Cloudera or with an authorized and
 *  properly licensed third party, you do not have any rights to this code.
 *
 *   If this code is provided to you under the terms of the AGPLv3:
 *   (A) CLOUDERA PROVIDES THIS CODE TO YOU WITHOUT WARRANTIES OF ANY KIND;
 *   (B) CLOUDERA DISCLAIMS ANY AND ALL EXPRESS AND IMPLIED WARRANTIES WITH RESPECT TO THIS CODE, INCLUDING BUT NOT
 *       LIMITED TO IMPLIED WARRANTIES OF TITLE, NON-INFRINGEMENT, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE;
 *   (C) CLOUDERA IS NOT LIABLE TO YOU, AND WILL NOT DEFEND, INDEMNIFY, OR HOLD YOU HARMLESS FOR ANY CLAIMS ARISING
 *       FROM OR RELATED TO THE CODE; AND
 *   (D) WITH RESPECT TO YOUR EXERCISE OF ANY RIGHTS GRANTED TO YOU FOR THE CODE, CLOUDERA IS NOT LIABLE FOR ANY
 *       DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, PUNITIVE OR CONSEQUENTIAL DAMAGES INCLUDING, BUT NOT LIMITED
 *       TO, DAMAGES RELATED TO LOST REVENUE, LOST PROFITS, LOSS OF INCOME, LOSS OF BUSINESS ADVANTAGE OR
 *       UNAVAILABILITY, OR LOSS OR CORRUPTION OF DATA.
 *
 */

package com.cloudera.naaf.download;

import com.cloudera.cdp.shaded.com.fasterxml.jackson.core.JsonProcessingException;
import com.cloudera.cdp.shaded.com.fasterxml.jackson.databind.ObjectMapper;
import com.cloudera.cdp.shaded.org.apache.commons.codec.binary.Base64;
import com.cloudera.cdp.shaded.org.bouncycastle.asn1.ASN1OctetString;
import com.cloudera.cdp.shaded.org.bouncycastle.asn1.pkcs.PrivateKeyInfo;
import com.cloudera.cdp.shaded.org.bouncycastle.crypto.params.Ed25519PrivateKeyParameters;
import com.cloudera.cdp.shaded.org.bouncycastle.crypto.signers.Ed25519Signer;

import java.io.IOException;
import java.nio.charset.StandardCharsets;
import java.security.InvalidKeyException;
import java.security.NoSuchAlgorithmException;
import java.security.PrivateKey;
import java.security.Signature;
import java.security.SignatureException;
import java.util.Objects;
import java.util.SortedMap;
import java.util.TreeMap;


/**
 * This code is copied directly from the com.cloudera.cdp.authentication.Signer class from the cdp-sdk-java module, in order to
 * reduce the library footprint.
 */
public class CdpSigner {

    private static final String RSA_ALGORITHM = "RSA";
    private static final String RSA_AUTH_METHOD = "rsav1";
    private static final String RSA_SIGNATURE_ALGORITHM = "SHA256withRSA";
    private static final String ED25519_ALGORITHM = "Ed25519";
    private static final String ED25519_AUTH_METHOD = "ed25519v1";
    private static final String ED25519_SIGNATURE_ALGORITHM = "Ed25519";

    /**
     * Computes the value for the x-altus-auth header for a request.
     * @param httpMethod the HTTP method
     * @param contentType the value of the content type header
     * @param date the value of the x-altus-date header
     * @param path the request portion of the URI, e.g. /iam/getUser
     * @param accessKeyId the access key ID
     * @param privateKey the access key private key
     * @return the value for the x-altus-auth header
     */
    public String computeAuthHeader(String httpMethod,
                                    String contentType,
                                    String date,
                                    String path,
                                    String accessKeyId,
                                    PrivateKey privateKey) {
        Objects.requireNonNull(httpMethod, "http method is required");
        Objects.requireNonNull(contentType, "content type is required");
        Objects.requireNonNull(date, "date is required");
        Objects.requireNonNull(path, "path is required");
        Objects.requireNonNull(accessKeyId, "access key id is required");
        Objects.requireNonNull(privateKey, "private key is required");

        String authMethod;
        String sigAlgo;
        switch (privateKey.getAlgorithm()) {
            case RSA_ALGORITHM:
                authMethod = RSA_AUTH_METHOD;
                sigAlgo = RSA_SIGNATURE_ALGORITHM;
                break;
            case ED25519_ALGORITHM:
                authMethod = ED25519_AUTH_METHOD;
                sigAlgo = ED25519_SIGNATURE_ALGORITHM;
                break;
            default:
                throw new RuntimeException("Unsupported key algorithm: " + privateKey.getAlgorithm());
        }
        byte[] bytesToSign = new StringBuilder()
                .append(httpMethod)
                .append("\n")
                .append(contentType)
                .append("\n")
                .append(date)
                .append("\n")
                .append(path)
                .append("\n")
                .append(authMethod)
                .toString()
                .getBytes(StandardCharsets.UTF_8);

        String signatureString;
        try {
            byte[] sig;
            switch (sigAlgo) {
                case RSA_SIGNATURE_ALGORITHM:
                    Signature signature = Signature.getInstance(sigAlgo);
                    signature.initSign(privateKey);
                    signature.update(bytesToSign);
                    sig = signature.sign();
                    break;
                case ED25519_SIGNATURE_ALGORITHM:
                    // This labourious process is necessary because we're trying to avoid
                    // relying on the BouncyCastle provider to be registered with the
                    // security subsystem. And the only way to get a Signature class is
                    // through the provider. The BC SignatureSpi implementation cannot
                    // be used directly.
                    PrivateKeyInfo info = PrivateKeyInfo.getInstance(privateKey.getEncoded());
                    ASN1OctetString asn1Seed = ASN1OctetString.getInstance(info.parsePrivateKey());
                    byte[] seed = asn1Seed.getOctets();

                    Ed25519Signer signer = new Ed25519Signer();
                    signer.init(true, new Ed25519PrivateKeyParameters(seed, 0));
                    signer.update(bytesToSign, 0, bytesToSign.length);
                    sig = signer.generateSignature();
                    break;
                default:
                    throw new RuntimeException("Unsupported key algorithm: " +
                            privateKey.getAlgorithm());
            }
            signatureString = Base64.encodeBase64URLSafeString(sig);
        } catch (NoSuchAlgorithmException |
                InvalidKeyException |
                SignatureException |
                IOException e) {
            throw new RuntimeException(e);
        }

        try {
            SortedMap<String, String> authParams = new TreeMap<>();
            authParams.put("access_key_id", accessKeyId);
            authParams.put("auth_method", authMethod);
            String encodedAuthParams =
                    new ObjectMapper().writeValueAsString(authParams);
            return String.format("%s.%s",
                    Base64.encodeBase64URLSafeString(
                            encodedAuthParams.getBytes(StandardCharsets.UTF_8)),
                    signatureString);
        } catch (JsonProcessingException e) {
            throw new RuntimeException("Error while building authentication header");
        }
    }
}
