/*
 * (c) 2018-2022 Cloudera, Inc. All rights reserved.
 *
 *   This code is provided to you pursuant to your written agreement with Cloudera, which may be the terms of the
 *  Affero General Public License version 3 (AGPLv3), or pursuant to a written agreement with a third party authorized
 *  to distribute this code.  If you do not have a written agreement with Cloudera or with an authorized and
 *  properly licensed third party, you do not have any rights to this code.
 *
 *   If this code is provided to you under the terms of the AGPLv3:
 *   (A) CLOUDERA PROVIDES THIS CODE TO YOU WITHOUT WARRANTIES OF ANY KIND;
 *   (B) CLOUDERA DISCLAIMS ANY AND ALL EXPRESS AND IMPLIED WARRANTIES WITH RESPECT TO THIS CODE, INCLUDING BUT NOT
 *       LIMITED TO IMPLIED WARRANTIES OF TITLE, NON-INFRINGEMENT, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE;
 *   (C) CLOUDERA IS NOT LIABLE TO YOU, AND WILL NOT DEFEND, INDEMNIFY, OR HOLD YOU HARMLESS FOR ANY CLAIMS ARISING
 *       FROM OR RELATED TO THE CODE; AND
 *   (D) WITH RESPECT TO YOUR EXERCISE OF ANY RIGHTS GRANTED TO YOU FOR THE CODE, CLOUDERA IS NOT LIABLE FOR ANY
 *       DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, PUNITIVE OR CONSEQUENTIAL DAMAGES INCLUDING, BUT NOT LIMITED
 *       TO, DAMAGES RELATED TO LOST REVENUE, LOST PROFITS, LOSS OF INCOME, LOSS OF BUSINESS ADVANTAGE OR
 *       UNAVAILABILITY, OR LOSS OR CORRUPTION OF DATA.
 *
 */

package com.cloudera.naaf.download;


import com.cloudera.cdp.shaded.org.bouncycastle.asn1.pkcs.PrivateKeyInfo;
import com.cloudera.cdp.shaded.org.bouncycastle.crypto.params.Ed25519PrivateKeyParameters;
import com.cloudera.cdp.shaded.org.bouncycastle.crypto.util.PrivateKeyInfoFactory;
import com.cloudera.cdp.shaded.org.bouncycastle.jcajce.provider.asymmetric.edec.KeyFactorySpi;
import com.cloudera.cdp.shaded.org.bouncycastle.util.io.pem.PemObject;
import com.cloudera.cdp.shaded.org.bouncycastle.util.io.pem.PemReader;

import java.io.IOException;
import java.io.StringReader;
import java.security.KeyFactory;
import java.security.NoSuchAlgorithmException;
import java.security.PrivateKey;
import java.security.spec.InvalidKeySpecException;
import java.security.spec.PKCS8EncodedKeySpec;
import java.util.Base64;
import java.util.Objects;

/**
 * This code is copied directly from com.cloudera.cdp.client.CredentialUtilities, in order to reduce the library footprint.
 */
public class CredentialUtilities {
    private static final int ED25519_KEY_LENGTH = 44;

    private CredentialUtilities() {
    }

    public static PrivateKey decodePrivateKey(final String privateKey) {
        Objects.requireNonNull(privateKey, "private key is required");
        return privateKey.length() == ED25519_KEY_LENGTH ? decodeEd25519PrivateKey(privateKey) : decodeRSAPrivateKey(privateKey);
    }

    private static PrivateKey decodeRSAPrivateKey(String privateKeyString) {
        privateKeyString = privateKeyString.replace("\\n", "\n");

        try {
            final PemReader pemReader = new PemReader(new StringReader(privateKeyString));
            Throwable t = null;

            PrivateKey privateKey;
            try {
                PemObject pemObject = pemReader.readPemObject();
                if (pemObject == null) {
                    throw new RuntimeException("Invalid private key ");
                }

                PKCS8EncodedKeySpec privateKeySpec = new PKCS8EncodedKeySpec(pemObject.getContent());
                KeyFactory factory = KeyFactory.getInstance("RSA");
                privateKey = factory.generatePrivate(privateKeySpec);
            } catch (final Throwable throwable) {
                t = throwable;
                throw throwable;
            } finally {
                if (pemReader != null) {
                    if (t != null) {
                        try {
                            pemReader.close();
                        } catch (final Throwable pemThrowable) {
                            t.addSuppressed(pemThrowable);
                        }
                    } else {
                        pemReader.close();
                    }
                }

            }

            return privateKey;
        } catch (final NoSuchAlgorithmException | InvalidKeySpecException | IOException e) {
            throw new RuntimeException("Unable to generate private key " + e.getMessage(), e);
        }
    }

    private static PrivateKey decodeEd25519PrivateKey(final String privateKey) {
        final byte[] seed = Base64.getDecoder().decode(privateKey);

        try {
            final Ed25519PrivateKeyParameters params = new Ed25519PrivateKeyParameters(seed, 0);
            final PrivateKeyInfo info = PrivateKeyInfoFactory.createPrivateKeyInfo(params);
            return (new KeyFactorySpi.Ed25519()).generatePrivate(info);
        } catch (final IOException e) {
            throw new RuntimeException("Unable to generate private key " + e.getMessage(), e);
        }
    }
}
