/*
 * (c) 2018-2022 Cloudera, Inc. All rights reserved.
 *
 *   This code is provided to you pursuant to your written agreement with Cloudera, which may be the terms of the
 *  Affero General Public License version 3 (AGPLv3), or pursuant to a written agreement with a third party authorized
 *  to distribute this code.  If you do not have a written agreement with Cloudera or with an authorized and
 *  properly licensed third party, you do not have any rights to this code.
 *
 *   If this code is provided to you under the terms of the AGPLv3:
 *   (A) CLOUDERA PROVIDES THIS CODE TO YOU WITHOUT WARRANTIES OF ANY KIND;
 *   (B) CLOUDERA DISCLAIMS ANY AND ALL EXPRESS AND IMPLIED WARRANTIES WITH RESPECT TO THIS CODE, INCLUDING BUT NOT
 *       LIMITED TO IMPLIED WARRANTIES OF TITLE, NON-INFRINGEMENT, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE;
 *   (C) CLOUDERA IS NOT LIABLE TO YOU, AND WILL NOT DEFEND, INDEMNIFY, OR HOLD YOU HARMLESS FOR ANY CLAIMS ARISING
 *       FROM OR RELATED TO THE CODE; AND
 *   (D) WITH RESPECT TO YOUR EXERCISE OF ANY RIGHTS GRANTED TO YOU FOR THE CODE, CLOUDERA IS NOT LIABLE FOR ANY
 *       DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, PUNITIVE OR CONSEQUENTIAL DAMAGES INCLUDING, BUT NOT LIMITED
 *       TO, DAMAGES RELATED TO LOST REVENUE, LOST PROFITS, LOSS OF INCOME, LOSS OF BUSINESS ADVANTAGE OR
 *       UNAVAILABILITY, OR LOSS OR CORRUPTION OF DATA.
 *
 */

package com.cloudera.naaf.download;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.ByteArrayOutputStream;
import java.io.Closeable;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.net.HttpURLConnection;
import java.net.URL;
import java.net.URLConnection;
import java.net.URLEncoder;
import java.security.PrivateKey;
import java.time.ZoneOffset;
import java.time.ZonedDateTime;
import java.time.format.DateTimeFormatter;
import java.util.Collections;
import java.util.HashMap;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

public class DfxFlowDownload implements FlowDownload {
    private static final Logger logger = LoggerFactory.getLogger(DfxFlowDownload.class);
    private static final int SUCCESS_HTTP_CODE = 200;
    private static final Pattern CRN_PATTERN = Pattern.compile("(crn:cdp:.*)/v\\.\\d+");

    private static final String CHAR_ENCODING = "UTF-8";
    private static final String JSON_CONTENT_TYPE = "application/json";
    private static final String GET_METHOD = "GET";
    private static final String DATE_HEADER_NAME = "x-altus-date";
    private static final String AUTH_HEADER_NAME = "x-altus-auth";

    private static final String CLIENT_TYPE = "function";

    private final String dfxGlobalBaseUrl;
    private final PrivateKey privateKey;
    private final String accessKey;

    public DfxFlowDownload(final String dfxGlobalBaseUrl, final String encodedPrivateKey, final String accessKey) {
        this.dfxGlobalBaseUrl = dfxGlobalBaseUrl;
        this.accessKey = accessKey;
        this.privateKey = CredentialUtilities.decodePrivateKey(encodedPrivateKey);
    }

    @Override
    public byte[] downloadFlow(final String flowVersionCrn) throws IOException {
        final String encodedVersionCrn = URLEncoder.encode(flowVersionCrn, CHAR_ENCODING);
        final String catalogDownloadPath = String.format("/dfx/api/v1/catalog/flows/versions/%s?clientType=%s", encodedVersionCrn, CLIENT_TYPE);
        final Response response = makeGetRequest(catalogDownloadPath);

        if (response.statusCode == SUCCESS_HTTP_CODE) {
            return response.content;
        } else {
            logger.error("Failed to fetch flow from Cloudera's DataFlow Service: received unsuccessful HTTP response " + response);

            final String responseText = new String(response.content);
            throw new IOException("Failed to fetch flow from Cloudera's DataFlow Service: received HTTP response code " + responseText);
        }
    }

    private String getFlowCrn(final String flowVersionCrn) {
        final Matcher matcher = CRN_PATTERN.matcher(flowVersionCrn);
        final boolean matches = matcher.matches();
        if (!matches) {
            return flowVersionCrn;
        }

        return matcher.group(1);
    }

    private Response makeGetRequest(final String urlPath) throws IOException {
        final String date = ZonedDateTime.now(ZoneOffset.UTC).format(DateTimeFormatter.RFC_1123_DATE_TIME);

        final CdpSigner signer = new CdpSigner();
        final String auth = signer.computeAuthHeader(GET_METHOD, JSON_CONTENT_TYPE, date, urlPath, accessKey, privateKey);

        final Map<String, String> headers = new HashMap<>();
        headers.put(DATE_HEADER_NAME, date);
        headers.put(AUTH_HEADER_NAME, auth);
        headers.put("Content-Type", JSON_CONTENT_TYPE);

        final String fullUrl;
        if (dfxGlobalBaseUrl.endsWith("/") && urlPath.startsWith("/")) {
            fullUrl = dfxGlobalBaseUrl + urlPath.substring(1);
        } else {
            fullUrl = dfxGlobalBaseUrl + urlPath;
        }

        return executeGetRequest(fullUrl, headers);
    }

    public static byte[] readResponse(final URLConnection urlConnection) throws IOException {
        final ByteArrayOutputStream buffer = new ByteArrayOutputStream();

        int bytesRead;
        byte[] data = new byte[256];

        try (final InputStream in = urlConnection.getInputStream()) {
            while ((bytesRead = in.read(data, 0, data.length)) != -1) {
                buffer.write(data, 0, bytesRead);
            }

            buffer.flush();
            return buffer.toByteArray();
        }
    }

    private static Response executeGetRequest(final String fullUrl, final Map<String, String> headers) throws IOException {
        final URL url = new URL(fullUrl);
        try (final HttpURLConnectionCloseable connectionCloseable = new HttpURLConnectionCloseable((HttpURLConnection) url.openConnection())) {
            final HttpURLConnection urlConnection = connectionCloseable.connection;
            urlConnection.setRequestMethod("GET");
            headers.forEach((key, value)  -> urlConnection.setRequestProperty(key, value));

            final int status = urlConnection.getResponseCode();

            return new Response(readResponse(urlConnection), status);
        } catch (final FileNotFoundException e) {
            logger.error("Failed to fetch flow from [{}]: not found", fullUrl);
            throw new IllegalArgumentException("Not found: " + fullUrl);
        }
    }

    // Temporary method to pull a flow file directly from a URL
    public static byte[] fetchDataflowFromUrl(final String flowUrl) throws IOException {
        final Response response = executeGetRequest(flowUrl, Collections.emptyMap());

        if (response.statusCode == SUCCESS_HTTP_CODE) {
            return response.content;
        } else {
            final String responseText = new String(response.content);
            logger.error("Failed to fetch flow from [{}]: received unsuccessful HTTP response {}", flowUrl, responseText);
            throw new IOException(responseText);
        }
    }

    private static class Response {
        private byte[] content;
        private int statusCode;

        public Response(byte[] content, int statusCode) {
            this.content = content;
            this.statusCode = statusCode;
        }
    }

    private static class HttpURLConnectionCloseable implements Closeable {
        private HttpURLConnection connection;

        private HttpURLConnectionCloseable(final HttpURLConnection connection) {
            this.connection = connection;
        }

        @Override
        public void close() {
            connection.disconnect();
        }
    }
}
