# NaaF: GCP Cloud Functions

NiFi as a Function - GCP Cloud Functions

This module provides the necessary components for running NiFi as a Function (NiFi Stateless) on Google Cloud Platform's Cloud Functions.


## Creating Cloud Function

To create a Cloud Function that is able to run NiFi as a Function, follow these steps.
1. Navigate to the Cloud Functions page (for example, https://console.cloud.google.com/functions/list)
2. Click the "Create function" button on the top-left of the page. 
3. Give the function a name and select the region. The Trigger settings vary depending on your use case, but be aware that
   only HTTP triggers can directly output data from the dataflow.  Click Save. 
4. Expand the "Runtime, build, connections and security settings".  Configure the Memory allocation and Timeout based on the discussion in the 
  [Runtime configuration](#runtime-configuration) section.
  Select a Runtime service account as discussed in the [Service Account](#service-account) section.  Configure the Runtime
  environment variables as discussed in the [Environment Variables](#configuring-environment-variables) section. 
5. You may also want to create secrets for sensitive parameters, as discussed in the [GCP Secret Manager](#gcp-secret-manager) section.
  These secrets can be mapped in the "SECURITY" tab of "Runtime, build, connections, and security settings". 
6. Click "Next"
7. For Runtime, select Java 11. In the Source code dropdown, select ZIP Upload, and upload the .zip file that was created from compiling the `naaf-gcp-assembly` module.
  Alternatively, you may choose "ZIP from Cloud Storage" and Browse to the Cloud Storage location of the .zip file. 
8. Set the Entry point to:
   1. HTTP triggers (synchronous): `com.cloudera.naaf.gcp.cloud.functions.StatelessNiFiHttpFunction`
   2. Background triggers (asynchronous): `com.cloudera.naaf.gcp.cloud.functions.StatelessNiFiBackgroundFunction`
9. Finally, click "Deploy".

Note: once the Cloud Function is deployed, it can take up to a few minutes to load the Edit Cloud Function page.  This is because
the GCP Console currently downloads the entire Source each time you edit the Cloud Function, which means approximately a 60MB download.

## Configuration

### Runtime configuration

In the Runtime Configuration tab, there are two very important configuration elements: the amount of memory
allocated to the Cloud Function and the Timeout. The appropriate values for these two settings will depend on the dataflow to be deployed. Typically, it is recommended
to start with a Memory allocation of 1GB - 2GB. This amount of memory may not be necessary, or more may be required, and this can be refined later.
Also note that any extensions or resources added to the Cloud Function as discussed in the [Cloud Storage](#cloud-storage) section 
will be placed under `/tmp`, which is an in-memory mount.  This means that they consume memory resources provisioned
for the function.

While NiFi as a Function tends to perform very well during a "Warm Start," a Cold Start that must source extensions may take several seconds to initialized. As a result,
it is recommended to set the Timeout to at least 30 seconds. If the dataflow to run reaches out to many other services or performs complex computation, it may be necessary
to use a larger value, even several minutes.
 

### Configuring Environment Variables

The function must now be configured to specify the dataflow to run and any necessary runtime configuration.
Configuration is provided using Environment Variables. On the "Runtime" tab, locate the "Runtime environment variables" section.

Click the "Add variable" button to add the first Environment variable.

The following Environment Variables are supported.

| Variable Name | Description | Required | Default Value |
| ------------- | ----------- | -------- | ------------- |
| FLOW_CRN | The Cloudera Resource Name (CRN) for the dataflow that is to be run. The dataflow must be stored in the DataFlow Service's Catalog. This CRN should indicate the specific version of the dataflow and as such will end with some suffix such as `/v.1` | true | -- |
| DF_PRIVATE_KEY | The Private Key to use for accessing Cloudera's DataFlow service. The Private Key and Access Key are used to authenticate with the DataFlow Service and must provide the necessary authorizations to access the specified dataflow. | true | -- |
| DF_ACCESS_KEY | The Access Key to use for accessing Cloudera's DataFlow service. The Private Key and Access Key are used to authenticate with the DataFlow Service and must provide the necessary authorizations to access the specified dataflow. | true | -- |
| INPUT_PORT | The name of the Input Port to use, if the dataflow has at least one.  If the dataflow has no Input Ports, this variable must not be specified.  If the specified dataflow has more than one Input Port at the root group level, this environment variable must be specified, indicating the name of the Input Port to queue up the Cloud Function event. If only one Input Port exists, this variable is not required but if specified must properly match the name of the Input Port. | false | -- |
| OUTPUT_PORT | The name of the Output Port to retrieve the results from. If no Output Port exists, this does not need to be specified and no data will be returned. If at least one Output Port exists in the dataflow, this variable can be used to determine the name of the Output Port whose data will be sent along as output of the function. Otherwise, see the section below on [Output Ports](#output-ports) to understand how the appropriate Output Port is determined. | false | -- |
| FAILURE_PORTS | A comma-separated list of Output Ports that exist at the root group level of the dataflow. If any FlowFile is sent to one of these Output Ports, the function invocation is considered a failure. See section below on [Output Ports](#output-ports) for more information. | false | -- |
| DF_SERVICE_URL | The Base URL for the Cloudera Dataflow Service. | false | https://api.us-west-1.cdp.cloudera.com |
| NEXUS_URL | The Base URL for a Nexus Repository to use for downloading any NiFi Archives (NARs) needed for running the dataflow. | false | https://repository.cloudera.com/artifactory/cloudera-repos/ |
| WORKING_DIR | The working directory, where NAR files will be expanded. | false | /tmp/working |
| EXTENSIONS_DOWNLOAD_DIR | The directory to which missing extensions / NiFi Archives (NARs) will be downloaded. See the section below on [Providing Custom Extensions / NARs](#providing-custom-extensions--nars) for more information. | false | /tmp/extensions |
| STORAGE_BUCKET | A Cloud Storage bucket in which to look in for custom extensions / NiFi Archives (NARs) and resources. See the section below on [Cloud Storage](#cloud-storage) for more information. | false | -- |
| STORAGE_EXTENSIONS_DIRECTORY | The directory in the Cloud Storage bucket to look in for custom extensions / NiFi Archives (NARs). See the section below on [Providing Custom Extensions / NARs](#providing-custom-extensions--nars) for more information. | false | extensions |
| STORAGE_RESOURCES_DIRECTORY | The directory in the Cloud Storage bucket to look in for custom resources. See the section below on [Providing Additional Resources](#providing-additional-resources) for more information. | false | resources |
| DISABLE_STATE_PROVIDER | If true, disables the Firestore dataflow state provider, even if the dataflow has stateful processors. | false | -- |
| FIRESTORE_STATE_COLLECTION | The Firestore collection name where the state will be stored | false | `nifi_state` |
| KRB5_FILE | Specifies the filename of the krb5.conf file. This is necessary only if connecting to a Kerberos-protected endpoint. See section below on configuring Kerberos. | false | `/etc/krb5.conf` |
| REGION | Specifies the region code of Google Cloud . For example, `us-west2` | false | -- |


Additionally, some Environment Variables apply only to HTTP triggers:

| Variable Name | Description | Required | Default Value |
| ------------- | ----------- | -------- | ------------- |
| HEADER_ATTRIBUTE_PATTERN | A RegEx capturing all flowfile attributes from the Output Port that should be returned as HTTP headers, if the trigger type is HTTP.  For all other trigger types, this variable is ignored. | false | -- |
| HTTP_STATUS_CODE_ATTRIBUTE | A flowfile attribute name that will set the response HTTP status code in a successful dataflow, if the trigger type is HTTP.  See the section below on [HTTP Triggers](#http-triggers) for more information.  For all other trigger types, this variable is ignored. | false | -- |

## Service Account

Cloud functions require a Service Account to use during runtime, granting the function permissions to use other GCP services.  The permissions
commonly required for NiFi as a Function will be discussed here.

When configuring your Cloud function, locate the "Runtime service account" section of the "Runtime, build, connections and security settings".
If you already have a Service Account for this purpose, you may select it here.  Otherwise, select "Create new service account".  This brings
up a sidebar allowing you to specify the account name and ID.  Once you click the "Create service account" button, you will need to add permissions
to the Service Account separately.

When configuring Secrets accessible to the Cloud function, as discussed in the [GCP Secret Manager](#gcp-secret-manager) section, the console
automatically prompts you to grant the necessary permissions to the Runtime Service Account.

If you are providing custom extensions or resources from a Cloud Storage bucket, as discussed in the [Cloud Storage](#cloud-storage) section, you
will need to add an additional role to the Service Account, as follows:
1. Navigate to the "IAM & Admin" GCP service, and select "Roles" on the left sidebar.
2. Click the "Create role" button on the top-left.
3. Name the role, and then click the "Add permissions" button.
4. In the Filter section, in place of "Enter property name or value", type "storage.buckets.get", and click the "Add" button.
5. Do the same for the "storage.buckets.list" permission.
6. Click the "Create" button.
7. Select "IAM" from the left sidebar.  Locate your Cloud function's Runtime Service Account, and click the Edit pencil icon on the far right.
8. In the Role dropdown menu, select "Custom" on the left side, and then your newly created Role on the right side.
9. Click "Save".  

Your Service Account now allows your Cloud function to list and get objects from buckets, but you now have to grant it access to the specific 
bucket configured in the `STORAGE_BUCKET` environment variable:
1. Navigate to the "Cloud Storage" GCP service, and click on your bucket name.
2. Select the "Permissions" tab underneath the bucket name.
3. In the Permissions section, click the "Add" button.
4. Type or paste the fully qualified name of the Runtime Service Account (e.g., `naaf@projectname.iam.gserviceaccount.com`).
5. In "Select a role", select "Cloud Storage" on the left, and "Storage Object Viewer" on the right.
6. Click the "Save" button.


## Output Ports

GCP Cloud functions may or may not allow the function to return a result.  The only GCP Cloud Function trigger type that can directly 
return a result is an HTTP trigger.  In this case, the output of the dataflow, as described below, is written as the HTTP response to the trigger.

### Background Triggers

With background functions, the Output Port dataflow logic is simpler.

### Background Triggers

With background functions, the Output Port dataflow logic is simpler.

If one Output Port is present, and its name is "failure", any data routed to that Output Port will cause the function invocation to fail.

If two Output Ports are present, and one of them is named "failure", that Output Port will behave as the "failure" port above.
Any other non-failure Output Port is considered a successful invocation, but the function will do nothing with data routed here.

Any Port with the name "failure" (ignoring case) will be considered a Failure Port. Additionally, if the "FAILURE_PORTS" environment variable
is specified, any port whose name is in the comma-separated list will be considered a Failure Port.


### HTTP Triggers

Since Cloud Functions with an HTTP trigger are invoked synchronously, they directly return a response.  The Output Port semantics
for this case is as follows.

If no Output Port is present in the dataflow's root group, no output will be provided from the function invocation, but a default
success message will be returned in the HTTP response with a status code of 200.

If one Output Port is present, and its name is "failure", any data routed to that Output Port will cause the function invocation to fail.
No output will be provided as the output of the function invocation, and an error message will be returned from the HTTP response,
with a status code of 500.

If two Output Ports are present, and one of them is named "failure", that Output Port will behave as the "failure" port above.
The other Output Port is considered a successful invocation. When the dataflow is invoked, if a single FlowFile is routed to this
Port, the contents of the FlowFile will be returned as the HTTP response. If more than one FlowFile is sent to the Output Port,
the dataflow will be considered a failure, as there is only one HTTP response possible. A MergeContent or MergeRecord
Processor may be used in order to assemble multiple FlowFiles into a single output FlowFile if necessary.

Any Port with the name "failure" (ignoring case) will be considered a Failure Port. Additionally, if the "FAILURE_PORTS" environment variable
is specified, any port whose name is in the comma-separated list will be considered a Failure Port.

If two or more Output Ports, other than failure ports, are present, the "OUTPUT_PORT" environment variable must be provided. In such a case,
this environment variable must match the name of an Output Port at the dataflow's root group (case is NOT ignored). Anything that is routed
to the specified port is considered the output of the function invocation. Anything routed to any other port is considered a failure.

Also note that in any successful response, the `mime.type` flowfile attribute will be used as the `Content-Type` of the response.  Conversely,
the HTTP request `Content-Type` will be added as the `mime.type` attribute of the input flowfile.

Additionally, there are two environment variables that can be configured to customize the HTTP response:
- `HEADER_ATTRIBUTE_PATTERN` can provide a regular expression matching any flowfile attributes to be included as HTTP response headers.
  This means that any attributes to be included as headers should be named as the headers themselves.
- `HTTP_STATUS_CODE_ATTRIBUTE` specifies a flowfile attribute that sets the HTTP status code in the response.  This allows a code other
  than 200 for 'success' or 500 for 'failure' to be returned.  If this environment variable is specified but the attribute
  is not set, the status code in the HTTP response will default to 200.

It is important to note that in the "failure" scenario, the `HTTP_STATUS_CODE_ATTRIBUTE` is not used.  Therefore, if a non-500 status code
is required, it must be provided in a successful dataflow path.  For example, assume a `409 Conflict` is desired in a particular case.
This could be achieved as follows:
1. In the dataflow, include an `UpdateAttribute` processor that sets an arbitrary attribute, say `http.status.code`, to the value `409`.
2. Optionally, add a `ReplaceText` processor that sets the contents of the flowfile using `Always Replace` to whatever HTTP
   response body is desired.
3. Route this part of the flow to the main Output Port used for the success case.
4. In the GCP Cloud function, set the Runtime environment variable `HTTP_STATUS_CODE_ATTRIBUTE` to `http.status.code`, and Deploy.

This same method may also be used to set a non-200 success code.

Finally, any HTTP request headers will appear in the dataflow as flowfile attributes, named `gcp.http.header.<HeaderName>`.

## Specifying Parameters

An important aspect of building a dataflow that is able to be run outside of the NiFi instance where the dataflow was built is the
concept of parameterization. NiFi allows users to build a dataflow making use of Parameter Contexts. This allows properties for
Processors and Controller Services to be defined at runtime instead of at build time.

Some parameters will be sensitive, such as a password, while others will not be, such as the name of a Kafka topic. The Cloud
function allows users to specify these parameters through use of Environment Variables, either directly or mapped from Secrets.


### Environment Variables

Any Parameter can be specified using the environment variables of the GCP Cloud function. When configuring the Cloud function,
in the "Runtime environment variables" section of the "Runtime, build, connections and security settings", simply add an environment variable 
whose name matches the name of a Parameter in your Parameter Context.


### GCP Secret Manager

A more secure mechanism for storing parameters is to use the GCP Secret Manager, which is recommended for sensitive properties of the dataflow. 
To do so, navigate to the "Secret Manager" GCP service and click the "Create secret" button at the top-left.  Name the secret descriptively
(this does not have to be the same name as a dataflow Parameter), and enter the Secret value (or upload a file, which
is most useful for mounted secrets as described below).  Finally, click the "Create secret" button at the bottom.  Repeat for all needed secrets.

Back in the "Runtime, build, connections and security settings" section of the Cloud function configuration, select the "Security" tab.
Now you can click "Reference a secret", which allows you to select an existing Secret.  At this point, the console will likely prompt
you to Grant permissions to use the secret to the Runtime Service Account, which you should do.

If your secret is an uploaded file, it probably makes the most sense to select "Mounted as volume" for the Reference method.

For parameters, select "Exposed as environment variable" as the Reference method.  Here you should enter the actual dataflow Parameter name
in the environment variable Name, since this is how parameters are mapped.  Optionally select "latest" from the "Version" dropdown menu, which
allows the secret value to be automatically updated if the function is redeployed.

It is important to note here that in order to use this capability, the Parameter Contexts used in the flow must have names that are compatible
with GCP Secret Manager names and Environment Variable names. To ensure that these naming conventions are followed, Parameter Contexts and
Parameters should use names that start with a letter and consist only of letters, numbers, and the underscore (_) character. For example,
instead of naming a Parameter "GCP Parameter Context" use a name of GCP_PARAMETER_CONTEXT when building the dataflow.

# Configuring Kerberos

Depending on the dataflow that is used, you may need to enable Kerberos authentication in order to interact with some service(s). To enable Kerberos, we must provide
an appropriate `krb5.conf` file. This file tells NaaF, among other things, where the Key Distribution Center (KDC) is located and the Kerberos Realm to use.
To specify this, the `krb5.conf` must be made available to NaaF. This can be done, for instance, using a GCS bucket.
See the section on [providing additional resources](#providing-additional-resources) to learn more.

Additionally, we must tell NiFi as a Function where it can find this `krb5.conf` file. To do this, add an environment variable to your function. The environment
variable must have the name `KRB5_FILE` and the value must be the (fully qualified) filename of the `krb5.conf` file. For example, `/tmp/resources/krb5.conf`.

In order to use Keytabs in your dataflow, it will also be important to provide the necessary keytab files to the NaaF function. This is accomplished in the same manner,
using a GCS Bucket. While the `krb5.conf` file is a system configuration, the keytab is not. The keytab will be configured in the dataflow on a per-component
basis. For example, one Processor may use keytab `nifi-keytab-1` while some Controller Service makes use of keytab `hive-keytab-4`. The location of these files
may change from deployment to deployment. For example, for an on-premises deployment, keytabs may be stored in `/etc/security/keytabs` while in NaaF, they may be made available in `/tmp/resources`.
As a result, it is advisable to use Parameter Contexts in order to parameterize the location of the keytab files. This is a good practice for any file-based resource that will
need to be made available to a deployment. In this case, because there is a parameter referencing the location, we need to provide NaaF with a value for that parameter. To do
so, we can simply create an environment variable (see [specifying parameters](#specifying-parameters) for more information).

# Dataflow State

By default, if your dataflow contains any stateful processors (e.g., `ListSFTP`), this state will automatically be stored in a Firestore collection called `nifi_state`,
though this name can be set using the `FIRESTORE_STATE_COLLECTION` environment variable.  In order to integrate with Firestore, some additional steps are required.
Without these steps, the dataflow will still run, but the state will not be preserved.

1. In the GCP console, navigate to the Firestore service.
2. Click the "Select Native Mode" button.
3. Select the appropriate region, and click the "Create database" button.
4. Navigate to the IAM service, and click the Edit (pencil) button on the service account principal that runs your Cloud Function.
5. Click the "Add Role" button, and add the "Firebase Admin" role, then Save.

The Firestore state provider can be disabled even if your dataflow contains stateful processors by setting the `DISABLE_STATE_PROVIDER` Environment Variable
to `true`.  Be warned, however, that this will cause processor state to be lost between function executions.

## Cloud Storage

### Providing Custom Extensions / NARs

The binary that is deployed for running the GCP Cloud function does not include the full NiFi deployment with all NARs / extensions. Doing so would result
in very long startup times and would require more expensive configurations for the Cloud function. Instead, it packages only the NARs necessary to run Stateless
NiFi. If any other extension is needed in order to run the dataflow, it will be automatically downloaded when the Cloud
function is initialized on a Cold Start. The extensions will be downloaded from the Cloudera Maven Repository by default.

However, an alternate Nexus repository can be configured by setting the `NEXUS_URL` environment variable to the URL of the Nexus Repository where
extensions should be downloaded. For example, to use Google's mirror of the Maven Central repository, set the `NEXUS_URL` environment variable to
`https://maven-central.storage-download.googleapis.com/maven2`. Any configured URL must either be accessible via `http` or be served over `https` with
a trusted certificate.

Typically, this is the desired approach for downloading extensions that are made available by Cloudera or Apache NiFi. However, if there is a need
to provide custom extensions, a simpler route is usually to use a Cloud Storage bucket. To do so, create a Cloud Storage bucket and create a directory
named descriptively, like "extensions".  Into this directory, upload all NAR files that will need to be used.

Then, we simply need to tell the Cloud function where to look for those extensions. We do this by adding another environment variable, `STORAGE_BUCKET` containing the name
of the bucket, and another called `STORAGE_EXTENSIONS_DIRECTORY` with the full directory path of the above extensions directory.  If the latter environment variable
is not specified, the directory defaults to "extensions".

Adding many extensions will cause the Cloud function to take a bit longer to load on a Cold Start, and will take more memory to load the additional
files and classes. Therefore, it is not recommended to upload hundreds of MBs or more of extensions if they are not needed.
However, this is supported if necessary. 


### Providing additional resources

In addition to NiFi extensions, some dataflows may also require additional resources in order to function. For example, a JDBC driver may be required for
establishing a database connection. Or a CSV file may be required in order to provide data enrichment. The recommended approach for providing such resources
to your Cloud function is via the Cloud Storage bucket.

Create a Cloud Storage bucket, if not already done as described in [Providing Custom Extensions / NARs](#providing-custom-extensions--nars) above, and create
a descriptively named directory, like "resources".  Into this directory, upload all resources required for the dataflow to run.  These resources will then
appear in the Cloud function's `/tmp/resources` directory for access by the dataflow.

Because each deployment of a given dataflow may need to reference files in a different location, depending on its environment, it is generally considered
a best practice to parameterize any resource that needs to be accessed by the dataflow. For example, if the dataflow contains a `DBCPConnectionPool` controller
service for accessing a database, it is recommended to use a Parameter for the "Database Driver Location(s)" property of the Controller Service. This allows each
deployment to specify the location of the JDBC Driver independently.

So, for example, we can set the "Database Driver Location(s)" property to a value of `#{JDBC_DRIVER_LOCATION}`. Then, in our Cloud Storage, we can upload a file that
contains our JDBC driver, `database-jdbc-driver.jar`. Then we can add an Environment Variable named `JDBC_DRIVER_LOCATION` with a value of `/tmp/resources/database-jdbc-driver.jar`.
By taking this approach, our dataflow is more reusable, as it can be deployed in many different environments.


## Testing the Dataflow

As with any dataflow that we build, it is important to be able to test our dataflow before attempting to deploy it. Any dataflow that is to be used by
GCP Cloud functions will require an Input Port and will most often contain at least one failure Output Port. This makes it quite easy to test our dataflows in NiFi!
We can simply move up to the Parent Process Group within NiFi and create a GenerateFlowFile processor. The "Custom Text" property can then be set to any
value that we may wish to feed into our dataflow. For example, to simulate an event indicating that data was added to a Cloud Storage bucket, we might set the "Custom Text"
property to the value:

```
{
  "name": "test/testfile.txt", 
  "bucket": "my-cloud-storage-bucket"
}
```

We can then have this GenerateFlowFile processor connected to the Input Port of our Process Group that we will run in the Cloud function. And the "Run Once" feature of NiFi
makes it really easy to create this FlowFile and send it to the Process Group. We can then start our Process Group and ensure that the data processes as expected.
If not, we can fix our dataflow and trigger the GenerateFlowFile processor again, until we have properly handled the data.

We can then Download the Flow Definition from the NiFi and upload the flow to the Cloudera DataFlow Service catalog.


## Testing the Deployment

Once we've built and verified our dataflow and uploaded it to the Cloudera DataFlow Service catalog, we can create our Cloud function. After deploying the function
as described in the first section, we will want to test that we have configured all of the settings. To do so, we can use the "Test" tab of the Cloud function to input the relevant JSON
structure.

We can then click on the "Test the function" button to have the dataflow run. If there are problems, we can easily inspect the log messages to help us to understand the
problem. If the logs do not provide enough details, we may need to adjust the log levels in order to gain more debugging information. See the section below
for more details.

The amount of time the cloud function takes to run depends heavily on the dataflow and how many extensions it needs. Because the Cloud function may have to download
some extensions from Nexus and perform initialization, a Cold Start may take many seconds. Approximately 20-30 seconds is not uncommon for a dataflow with
several extensions, while other dataflows may complete in 10 seconds. After we obtain a successful run of our function using the Test button, it may be helpful
to run several additional iterations in order to understand how the function will perform when a Cold Start is not necessary. Again, this depends heavily upon
the configured dataflow and other services that it may interact with. A simple dataflow may complete in as little as 10 milliseconds while a dataflow that must
perform complex transformations and reach out to one or more web services or databases for enrichment may take several seconds to complete.

It will be important to adjust Cloud function's Timeout configuration in the "Runtime" tab of the "Runtime, build, connections and security settings" if the Cold Start takes longer
than the amount of time allocated. 


## Adjusting Log Levels

NiFi as a Function makes use of the SLF4J Simple Logger for logging purposes. As such, we can update the log levels by adjusting the JVM's system properties.
To do so, we set the `JAVA_TOOL_OPTIONS` environment variable. To set a log level for a given logger, we set the JVM system property named 
`org.slf4j.simpleLogger.log.<logger name>` to the desired log level. For example, if we want to enable DEBUG logging for the `StandardStatelessNiFiFunction`
in order to view the actual data flow JSON being used, we would set the `JAVA_TOOL_OPTIONS` environment variable to
`-Dorg.slf4j.simpleLogger.log.org.com.cloudera.naaf.StandardStatelessNiFiFunction=DEBUG`.

We can also set additional logger levels by adding multiple `-D` options separated by spaces. For example, to enable DEBUG logs on both the
NiFi as a Function framework and the Stateless Bootstrap class, we would set `JAVA_TOOL_OPTIONS` to a value of
`-Dorg.slf4j.simpleLogger.log.com.cloudera.naaf.StandardStatelessNiFiFunction=DEBUG -Dorg.slf4j.simpleLogger.log.org.apache.nifi.stateless.bootstrap.StatelessBootstrap=DEBUG`


## Handling Failures

As we mentioned, we can use a `failure` Output Port to notify the Cloud function that an invocation has failed. At that point, the Cloud function may automatically retry the invocation,
if configured in the "Trigger" section, for up to seven days. After that, if it still fails, the Cloud function will simply drop the event notification. This could result in data loss, as it means
that our Cloud function will not have a chance to process the data again.
