/*
 *  Copyright (c) 2018-2021 Cloudera, Inc. All rights reserved.
 *
 *  This code is provided to you pursuant to your written agreement with Cloudera, which may be the terms of the
 *  Affero General Public License version 3 (AGPLv3), or pursuant to a written agreement with a third party authorized
 *  to distribute this code.  If you do not have a written agreement with Cloudera or with an authorized and
 *  properly licensed third party, you do not have any rights to this code.
 *
 *  If this code is provided to you under the terms of the AGPLv3:
 *   (A) CLOUDERA PROVIDES THIS CODE TO YOU WITHOUT WARRANTIES OF ANY KIND;
 *   (B) CLOUDERA DISCLAIMS ANY AND ALL EXPRESS AND IMPLIED WARRANTIES WITH RESPECT TO THIS CODE, INCLUDING BUT NOT
 *       LIMITED TO IMPLIED WARRANTIES OF TITLE, NON-INFRINGEMENT, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE;
 *   (C) CLOUDERA IS NOT LIABLE TO YOU, AND WILL NOT DEFEND, INDEMNIFY, OR HOLD YOU HARMLESS FOR ANY CLAIMS ARISING
 *       FROM OR RELATED TO THE CODE; AND
 *   (D) WITH RESPECT TO YOUR EXERCISE OF ANY RIGHTS GRANTED TO YOU FOR THE CODE, CLOUDERA IS NOT LIABLE FOR ANY
 *       DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, PUNITIVE OR CONSEQUENTIAL DAMAGES INCLUDING, BUT NOT LIMITED
 *       TO, DAMAGES RELATED TO LOST REVENUE, LOST PROFITS, LOSS OF INCOME, LOSS OF BUSINESS ADVANTAGE OR
 *       UNAVAILABILITY, OR LOSS OR CORRUPTION OF DATA.
 *
 */

package com.cloudera.naaf.gcp.cloud.functions;

import com.cloudera.naaf.StatelessNiFiFunction;
import com.cloudera.naaf.StatelessNiFiUtil;
import com.google.cloud.functions.Context;
import com.google.cloud.functions.RawBackgroundFunction;
import org.apache.nifi.stateless.flow.TriggerResult;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.nio.charset.StandardCharsets;
import java.util.HashMap;
import java.util.Map;
import java.util.concurrent.atomic.AtomicReference;

public class StatelessNiFiBackgroundFunction implements RawBackgroundFunction {
    private static final Logger logger = LoggerFactory.getLogger(StatelessNiFiBackgroundFunction.class);
    private static final String ATTRIBUTE_PREFIX = "gcp.context.";

    private static final AtomicReference<StatelessNiFiGcpFunctionHolder> functionHolderReference = new AtomicReference<>();


    @Override
    public void accept(final String input, final Context context) throws Exception {
        final StatelessNiFiGcpFunctionHolder functionHolder = getFunctionHolder();

        logger.debug("Triggering background function...");
        final Map<String, String> attributes = getAttributes(context);
        try (final InputStream bais = new ByteArrayInputStream(input.getBytes(StandardCharsets.UTF_8))) {
            final String instanceId = functionHolder.getFunctionName();
            final TriggerResult triggerResult = functionHolder.getStatelessNiFiFunction().trigger(instanceId, bais, attributes);
            StatelessNiFiUtil.handleTriggerResult(triggerResult);
        }
    }

    private StatelessNiFiGcpFunctionHolder getFunctionHolder() throws IOException {
        StatelessNiFiGcpFunctionHolder functionHolder = functionHolderReference.get();
        if (functionHolder == null) {
            logger.debug("Function has not been initialized yet, initializing now...");
            functionHolder = new StatelessNiFiGcpFunctionHolder();
            functionHolderReference.set(functionHolder);
            logger.debug("Successfully initialized Background Function");

            return functionHolder;
        }

        logger.debug("Function is already initialized");
        return functionHolder;
    }

    private Map<String, String> getAttributes(final Context context) {
        final Map<String, String> attributes = new HashMap<>();
        if (context.attributes() != null) {
            context.attributes().forEach((key, value) -> attributes.put(ATTRIBUTE_PREFIX + "attributes." + key, value));
        }

        attributes.put(StatelessNiFiFunction.CLOUD_PROVIDER_ATTRIBUTE, StatelessNiFiGcpFunctionHolder.GCP_CLOUD_PROVIDER);
        attributes.put(ATTRIBUTE_PREFIX + "eventId", context.eventId());
        attributes.put(ATTRIBUTE_PREFIX + "eventType", context.eventType());
        attributes.put(ATTRIBUTE_PREFIX + "resource", context.resource());
        attributes.put(ATTRIBUTE_PREFIX + "timestamp", context.timestamp());
        return attributes;
    }
}
