/*
 *  Copyright (c) 2018-2021 Cloudera, Inc. All rights reserved.
 *
 *  This code is provided to you pursuant to your written agreement with Cloudera, which may be the terms of the
 *  Affero General Public License version 3 (AGPLv3), or pursuant to a written agreement with a third party authorized
 *  to distribute this code.  If you do not have a written agreement with Cloudera or with an authorized and
 *  properly licensed third party, you do not have any rights to this code.
 *
 *  If this code is provided to you under the terms of the AGPLv3:
 *   (A) CLOUDERA PROVIDES THIS CODE TO YOU WITHOUT WARRANTIES OF ANY KIND;
 *   (B) CLOUDERA DISCLAIMS ANY AND ALL EXPRESS AND IMPLIED WARRANTIES WITH RESPECT TO THIS CODE, INCLUDING BUT NOT
 *       LIMITED TO IMPLIED WARRANTIES OF TITLE, NON-INFRINGEMENT, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE;
 *   (C) CLOUDERA IS NOT LIABLE TO YOU, AND WILL NOT DEFEND, INDEMNIFY, OR HOLD YOU HARMLESS FOR ANY CLAIMS ARISING
 *       FROM OR RELATED TO THE CODE; AND
 *   (D) WITH RESPECT TO YOUR EXERCISE OF ANY RIGHTS GRANTED TO YOU FOR THE CODE, CLOUDERA IS NOT LIABLE FOR ANY
 *       DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, PUNITIVE OR CONSEQUENTIAL DAMAGES INCLUDING, BUT NOT LIMITED
 *       TO, DAMAGES RELATED TO LOST REVENUE, LOST PROFITS, LOSS OF INCOME, LOSS OF BUSINESS ADVANTAGE OR
 *       UNAVAILABILITY, OR LOSS OR CORRUPTION OF DATA.
 *
 */

package com.cloudera.naaf.gcp.cloud.functions;

import com.cloudera.naaf.StandardStatelessNiFiFunction;
import com.cloudera.naaf.StatelessFunctionConfiguration;
import com.cloudera.naaf.StatelessNiFiFunction;
import com.cloudera.naaf.metering.CloudPlatform;

import java.io.File;
import java.io.IOException;
import java.util.Optional;

public class StatelessNiFiGcpFunctionHolder {
    private static final File RESOURCES_DIRECTORY = new File(StatelessNiFiFunction.TEMP_DIR, "resources");

    // Environment Variables
    private static final String FUNCTION_NAME_ENV_VAR = "K_SERVICE";
    private static final String FUNCTION_VERSION_ENV_VAR = "K_REVISION";
    private static final String STORAGE_EXTENSIONS_DIRECTORY_ENV_VAR = "STORAGE_EXTENSIONS_DIRECTORY";
    private static final String STORAGE_RESOURCES_DIRECTORY_ENV_VAR = "STORAGE_RESOURCES_DIRECTORY";
    private static final String STORAGE_BUCKET_ENV_VAR = "STORAGE_BUCKET";
    private static final String RESOURCE_DOWNLOADER_TYPE = "com.cloudera.naaf.gcp.cloud.functions.resources.GcpCloudStorageResourceDownloader";
    private static final String STATE_PROVIDER_TYPE = "com.cloudera.naaf.gcp.cloud.functions.FirestoreStatelessNiFiStateProvider";
    private static final String REGION = "REGION";

    // Defaults
    private static final String DEFAULT_STORAGE_EXTENSIONS_DIRECTORY = "extensions";
    private static final String DEFAULT_STORAGE_RESOURCES_DIRECTORY = "resources";

    static final String GCP_CLOUD_PROVIDER = "GCP";

    private final StatelessNiFiFunction statelessNiFiFunction;

    private final String functionName;

    public StatelessNiFiGcpFunctionHolder() throws IOException {
        functionName = System.getenv(FUNCTION_NAME_ENV_VAR);
        final String functionVersion = System.getenv(FUNCTION_VERSION_ENV_VAR);
        final String region = System.getenv(REGION);

        final File resourcesDirectory = RESOURCES_DIRECTORY;
        if (!resourcesDirectory.exists() && !resourcesDirectory.mkdirs()) {
            throw new IOException("Resources directory " + resourcesDirectory.getAbsolutePath() + " does not exist and cannot be created");
        }
        String extensionsBucketDirectory = null;
        String resourcesBucketDirectory = null;

        final String bucketName = System.getenv(STORAGE_BUCKET_ENV_VAR);

        if (bucketName != null) {
            extensionsBucketDirectory = Optional.ofNullable(System.getenv(STORAGE_EXTENSIONS_DIRECTORY_ENV_VAR))
                    .orElse(DEFAULT_STORAGE_EXTENSIONS_DIRECTORY);
            resourcesBucketDirectory = Optional.ofNullable(System.getenv(STORAGE_RESOURCES_DIRECTORY_ENV_VAR))
                    .orElse(DEFAULT_STORAGE_RESOURCES_DIRECTORY);
        }

        final StatelessFunctionConfiguration configurationActions = new StatelessFunctionConfiguration.Builder()
                .systemClassLoader(this.getClass().getClassLoader())
                .cloudPlatform(CloudPlatform.GOOGLE_CLOUD_FUNCTIONS)
                .resourcesDirectory(resourcesDirectory)
                .resourcesSourceDirectory(resourcesBucketDirectory)
                .extensionsSourceDirectory(extensionsBucketDirectory)
                .resourceDownloaderType(RESOURCE_DOWNLOADER_TYPE)
                .stateProviderType(STATE_PROVIDER_TYPE)
                .region(region)
                .build();
        final StandardStatelessNiFiFunction statelessNiFiFunction = new StandardStatelessNiFiFunction(functionName, functionVersion, configurationActions);

        this.statelessNiFiFunction = statelessNiFiFunction;
    }

    /**
     * Returns the Stateless NiFi function, which can be used to trigger the function.
     * @return The Stateless NiFi function
     */
    public StatelessNiFiFunction getStatelessNiFiFunction() {
        return statelessNiFiFunction;
    }

    public void finalize() {
        statelessNiFiFunction.cleanUp();
    }

    String getFunctionName() {
        return functionName;
    }
}
